# -*- coding: utf-8 -*-
"""Base Site class."""
from typing import ClassVar
from typing import Dict

from mrsimulator.utils.parseable import Parseable
from pydantic import validator

from .isotope import Isotope
from .tensors import AntisymmetricTensor
from .tensors import SymmetricTensor

__author__ = "Deepansh Srivastava"
__email__ = "srivastava.89@osu.edu"


class Site(Parseable):
    """Base class representing a single-site nuclear spin interaction tensor parameters.
    The single-site nuclear spin interaction tensors include the nuclear shielding
    and the electric quadrupolar tensor.

    .. rubric:: Attribute Documentation

    Attributes
    ----------

    isotope: str (optional).
        A string expressed as an atomic number followed by an isotope symbol, eg.,
        `'13C'`, `'17O'`. The default value is `'1H'`.

        Example
        -------

        >>> site = Site(isotope='2H')

    isotropic_chemical_shift: float (optional).
        The isotropic chemical shift of the site in ppm. The default value is 0.

        Example
        -------

        >>> site.isotropic_chemical_shift = 43.3

    shielding_symmetric: :ref:`sy_api` or equivalent dict object (optional).
        The attribute represents the parameters of the irreducible second-rank traceless
        symmetric part of the nuclear shielding tensor. The default value is None.

        The allowed attributes of the :ref:`sy_api` class for `shielding_symmetric` are
        ``zeta``, ``eta``, ``alpha``, ``beta``, and ``gamma``, where ``zeta`` is the
        shielding anisotropy, in ppm, and ``eta`` is the shielding asymmetry parameter
        defined using the Haeberlen convention. The Euler angles ``alpha``, ``beta``,
        and ``gamma`` are in radians.

        Example
        -------

        >>> site.shielding_symmetric = {'zeta': 10, 'eta': 0.5}

        >>> # or equivalently
        >>> site.shielding_symmetric = SymmetricTensor(zeta=10, eta=0.5)

    shielding_antisymmetric: :ref:`asy_api` or equivalent dict object (optional).
        The attribute represents the parameters of the irreducible first-rank
        antisymmetric part of the nuclear shielding tensor. The default value is None.

        The allowed attributes of the :ref:`asy_api` class for
        `shielding_antisymmetric` are ``zeta``, ``alpha``, and ``beta``, where ``zeta``
        is the anisotropy parameter, in ppm, of the anti-symmetric first-rank tensor.
        The angles ``alpha`` and ``beta`` are in radians.

        Example
        -------

        >>> site.shielding_antisymmetric = {'zeta': 20}

        >>> # or equivalently
        >>> site.shielding_antisymmetric = AntisymmetricTensor(zeta=20)

    quadrupolar: :ref:`sy_api` or equivalent dict object (optional).
        The attribute represents the parameters of the traceless irreducible second-rank
        symmetric part of the electric-field gradient tensor. The default value is None.

        The allowed attributes of the :ref:`sy_api` class for `quadrupolar` are ``Cq``,
        ``eta``, ``alpha``, ``beta``, and ``gamma``, where ``Cq`` is the quadrupolar
        coupling constant, in Hz, and ``eta`` is the quadrupolar asymmetry parameter.
        The Euler angles ``alpha``, ``beta``, and ``gamma`` are in radians.

        Example
        -------

        >>> site.quadrupolar = {'Cq': 3.2e6, 'eta': 0.52}

        >>> # or equivalently
        >>> site.quadrupolar = SymmetricTensor(Cq=3.2e6, eta=0.52)

    name: str (optional).
        The name or id of the site. The default value is None.

        Example
        -------

        >>> site.name = '2H-0'
        >>> site.name
        '2H-0'

    label: str (optional).
        The label for the site. The default value is None.

        Example
        -------

        >>> site.label = 'Quad site'
        >>> site.label
        'Quad site'

    description: str (optional).
        A description of the site. The default value is None.

        Example
        -------

        >>> site.description = 'An example Quadrupolar site.'
        >>> site.description
        'An example Quadrupolar site.'

    Example
    -------

    The following are a few examples of the site object.

    >>> site1 = Site(
    ...     isotope='33S',
    ...     isotropic_chemical_shift=20, # in ppm
    ...     shielding_symmetric={
    ...         "zeta": 10, # in ppm
    ...         "eta": 0.5
    ...     },
    ...     quadrupolar={
    ...         "Cq": 5.1e6, # in Hz
    ...         "eta": 0.5
    ...     }
    ... )

    Using SymmetricTensor objects.

    >>> site1 = Site(
    ...     isotope='13C',
    ...     isotropic_chemical_shift=20, # in ppm
    ...     shielding_symmetric=SymmetricTensor(zeta=10, eta=0.5),
    ... )
    """

    name: str = None
    label: str = None
    description: str = None
    isotope: str = "1H"
    isotropic_chemical_shift: float = 0.0
    shielding_symmetric: SymmetricTensor = None
    shielding_antisymmetric: AntisymmetricTensor = None
    quadrupolar: SymmetricTensor = None

    property_unit_types: ClassVar = {"isotropic_chemical_shift": "dimensionless"}
    property_default_units: ClassVar = {"isotropic_chemical_shift": "ppm"}
    property_units: Dict = {"isotropic_chemical_shift": "ppm"}

    @validator("quadrupolar")
    def spin_must_be_at_least_one(cls, v, values):
        if v is None:
            return v
        isotope = values["isotope"]
        isotope = Isotope(**isotope) if isinstance(isotope, dict) else isotope
        spin_I = isotope.spin
        if spin_I < 1:
            message = (
                f"with spin quantum number {spin_I} does not allow quadrupolar tensor."
            )
            raise ValueError(f"{isotope} {message}")
        _ = [
            v.property_units.pop(item) if item in v.property_units else None
            for item in ["D", "zeta"]
        ]
        return v

    @validator("shielding_symmetric", "shielding_antisymmetric")
    def shielding_symmetric_must_not_contain_Cq_and_D(cls, v, values):
        _ = [
            v.property_units.pop(item) if item in v.property_units else None
            for item in ["Cq", "D"]
        ]
        return v

    @validator("isotope", always=True)
    def validate_isotope(cls, v, *, values, **kwargs):
        return Isotope(symbol=v)

    class Config:
        validate_assignment = True

    @classmethod
    def parse_dict_with_units(cls, py_dict: dict):
        """Parse the physical quantity from a dictionary representation of the Site
        object, where the physical quantity is expressed as a string with a number and
        a unit.

        Args:
            dict py_dict: A required python dict object.

        Returns:
            :ref:`site_api` object.

        Example
        -------

        >>> site_dict = {
        ...    "isotope": "13C",
        ...    "isotropic_chemical_shift": "20 ppm",
        ...    "shielding_symmetric": {"zeta": "10 ppm", "eta":0.5}
        ... }
        >>> site1 = Site.parse_dict_with_units(site_dict)
        """
        prop_mapping = {
            "shielding_symmetric": SymmetricTensor,
            "shielding_antisymmetric": AntisymmetricTensor,
            "quadrupolar": SymmetricTensor,
        }

        for k, v in prop_mapping.items():
            if k in py_dict:
                py_dict[k] = v.parse_dict_with_units(py_dict[k])

        return super().parse_dict_with_units(py_dict)

    # Deprecated
    # def to_freq_dict(self, B0):
    #     """
    #     Serialize the Site object to a JSON compliant python dictionary object, where
    #     the attribute value is a number expressed in the attribute's default unit.
    #     The default unit for the attributes with respective dimensionalities is:

    #     - frequency: ``Hz``
    #     - angle: ``rad``

    #     Args:
    #         float B0: A required macroscopic magnetic flux density in units of T.

    #     Return:
    #         Python dict object.

    #     Example
    #     -------

    #     >>> pprint(site1.to_freq_dict(B0=9.4))
    #     {'description': None,
    #      'isotope': '13C',
    #      'isotropic_chemical_shift': -2013.1791999999998,
    #      'label': None,
    #      'name': None,
    #      'quadrupolar': None,
    #      'shielding_antisymmetric': None,
    #      'shielding_symmetric': {'alpha': None,
    #                              'beta': None,
    #                              'eta': 0.5,
    #                              'gamma': None,
    #                              'zeta': -1006.5895999999999}}
    #     """
    #     temp_dict = self.dict(exclude={"isotope"})
    #     temp_dict["isotope"] = self.isotope.symbol
    #     larmor_frequency = -self.isotope.gyromagnetic_ratio * B0  # in MHz
    #     for k in ["shielding_symmetric", "shielding_antisymmetric", "quadrupolar"]:
    #         if getattr(self, k):
    #             temp_dict[k] = getattr(self, k).to_freq_dict(larmor_frequency)
    #             if k == "shielding_symmetric":
    #                 temp_dict[k].pop("Cq")

    #     temp_dict["isotropic_chemical_shift"] *= larmor_frequency
    #     temp_dict.pop("property_units")
    #     return temp_dict



# -*- coding: utf-8 -*-
from .base import BlochDecayCentralTransitionSpectrum  # noqa:F401
from .base import BlochDecayCTSpectrum  # noqa:F401
from .base import BlochDecaySpectrum  # noqa:F401
from .base import Method1D  # noqa:F401
from .base import Method2D  # noqa:F401
from .mqvas import FiveQ_VAS  # noqa:F401
from .mqvas import SevenQ_VAS  # noqa:F401
from .mqvas import ThreeQ_VAS  # noqa:F401
from .ssb2d import SSB2D  # noqa:F401
from .stvas import ST1_VAS  # noqa:F401
from .stvas import ST2_VAS  # noqa:F401

# from .correlation import Cosy  # noqa:F401
# from .correlation import Inadequate  # noqa:F401


__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"



# -*- coding: utf-8 -*-
import numpy as np
from mrsimulator import Site
from mrsimulator import SpinSystem

__author__ = "Deepansh Srivastava"
__email__ = "srivastava.89@osu.edu"


SHIELDING_SYM_PARAMS = ["zeta", "eta", "alpha", "beta", "gamma"]
QUADRUPOLAR_PARAMS = ["Cq", "eta", "alpha", "beta", "gamma"]


def single_site_system_generator(
    isotopes,
    isotropic_chemical_shifts=0,
    shielding_symmetric=None,
    quadrupolar=None,
    abundance=None,
    site_labels=None,
    site_names=None,
    site_descriptions=None,
    rtol=1e-3,
):
    r"""Generate and return a list of single-site spin systems from the input parameters.

    Args:

        isotopes:
            A string or a list of site isotopes.
        isotropic_chemical_shifts:
            A float or a list/ndarray of values. The default value is 0.
        shielding_symmetric:
            A shielding symmetric like dict object, where the keyword value can either
            be a float or a list/ndarray of floats. The default value is None. The
            allowed keywords are ``zeta``, ``eta``, ``alpha``, ``beta``, and ``gamma``.
        quadrupolar:
            A quadrupolar like dict object, where the keyword value can either be a
            float or a list/ndarray of floats. The default value is None. The allowed
            keywords are ``Cq``, ``eta``, ``alpha``, ``beta``, and ``gamma``.
        abundance:
            A float or a list/ndarray of floats describing the abundance of each spin
            system.
        site_labels:
            A string or a list of strings each with a site label. The default is None.
        site_names:
            A string or a list of strings each with a site name. The default is None.
        site_descriptions:
            A string or a list of strings each with a site description. Default is None.
        rtol:
            The relative tolerance. This value is used in determining the cutoff
            abundance given as
            :math:`\tt{abundance}_\tt{cutoff} = \tt{rtol} * \tt{max(abundance)}.`
            The spin systems with abundance below this threshold are ignored.

    .. note::
        The parameter value can either be a float or a list/ndarray. If the parameter
        value is a float, the given value is assigned to the respective parameter in all
        the spin systems. If the parameter value is a list or ndarray, its ith value is
        assigned to the respective parameter of the ith spin system. When multiple
        parameter values are given as lists/ndarrays, the length of all the lists must
        be the same.
    """
    isotopes = _fix_item(isotopes)
    n_isotopes = _get_length(isotopes)

    isotropic_chemical_shifts = _fix_item(isotropic_chemical_shifts)
    n_iso = _get_length(isotropic_chemical_shifts)

    abundance = _fix_item(abundance)
    n_abd = _get_length(abundance)

    site_labels = _fix_item(site_labels)
    n_site_labels = _get_length(site_labels)

    site_names = _fix_item(site_names)
    n_site_names = _get_length(site_names)

    site_descriptions = _fix_item(site_descriptions)
    n_site_descriptions = _get_length(site_descriptions)

    n_ss, shield_keys, shielding_symmetric = _get_shielding_info(shielding_symmetric)
    n_q, quad_keys, quadrupolar = _get_quad_info(quadrupolar)

    n_len = _check_size(
        np.asarray(
            [
                n_isotopes,
                n_iso,
                *n_ss,
                *n_q,
                n_abd,
                n_site_labels,
                n_site_names,
                n_site_descriptions,
            ]
        )
    )

    if abundance is None:
        abundance = 1 / n_len

    # system with only isotope and isotropic shifts parameters
    isotopes_ = _get_default_lists(isotopes, n_len)
    iso_chemical_shifts_ = _get_default_lists(isotropic_chemical_shifts, n_len)
    abundance_ = _get_default_lists(abundance, n_len)
    site_labels_ = _get_default_lists(site_labels, n_len)
    site_names_ = _get_default_lists(site_names, n_len)
    site_descriptions_ = _get_default_lists(site_descriptions, n_len)

    index = np.where(abundance_ > rtol * abundance_.max())[0]

    sys = [
        SpinSystem(
            sites=[
                Site(
                    isotope=ist__,
                    isotropic_chemical_shift=iso__,
                    label=lab__,
                    name=name__,
                    description=dis__,
                )
            ],
            abundance=abd__,
        )
        for ist__, iso__, lab__, name__, dis__, abd__ in zip(
            isotopes_[index],
            iso_chemical_shifts_[index],
            site_labels_[index],
            site_names_[index],
            site_descriptions_[index],
            abundance_[index],
        )
    ]

    # system with additional shielding symmetric parameters
    if shielding_symmetric is not None:
        lst = [
            _get_default_lists(shielding_symmetric[item], n_len)[index]
            if item in shield_keys
            else [None for _ in range(index.size)]
            for item in SHIELDING_SYM_PARAMS
        ]
        _populate_shielding(sys, lst)

    # system with additional quadrupolar parameters
    if quadrupolar is not None:
        lst = [
            _get_default_lists(quadrupolar[item], n_len)[index]
            if item in quad_keys
            else [None for _ in range(index.size)]
            for item in QUADRUPOLAR_PARAMS
        ]
        _populate_quadrupolar(sys, lst)

    return sys


def _get_shielding_info(shielding_symmetric):
    n_ss, shield_keys = [], []
    if shielding_symmetric is not None:
        shield_keys = shielding_symmetric.keys()
        shielding_symmetric = {
            item: _fix_item(shielding_symmetric[item])
            for item in SHIELDING_SYM_PARAMS
            if item in shield_keys
        }
        n_ss = [
            _get_length(shielding_symmetric[item])
            for item in SHIELDING_SYM_PARAMS
            if item in shield_keys
        ]
    return n_ss, shield_keys, shielding_symmetric


def _get_quad_info(quadrupolar):
    n_q, quad_keys = [], []
    if quadrupolar is not None:
        quad_keys = quadrupolar.keys()
        quadrupolar = {
            item: _fix_item(quadrupolar[item])
            for item in QUADRUPOLAR_PARAMS
            if item in quad_keys
        }
        n_q = [
            _get_length(quadrupolar[item])
            for item in QUADRUPOLAR_PARAMS
            if item in quad_keys
        ]
    return n_q, quad_keys, quadrupolar


def _populate_quadrupolar(sys, items):
    n = len(items[0])
    for i in range(n):
        if sys[i].sites[0].isotope.spin > 0.5:
            sys[i].sites[0].quadrupolar = {
                "Cq": items[0][i],
                "eta": items[1][i],
                "alpha": items[2][i],
                "beta": items[3][i],
                "gamma": items[4][i],
            }


def _populate_shielding(sys, items):
    n = len(items[0])
    for i in range(n):
        sys[i].sites[0].shielding_symmetric = {
            "zeta": items[0][i],
            "eta": items[1][i],
            "alpha": items[2][i],
            "beta": items[3][i],
            "gamma": items[4][i],
        }


def _get_default_lists(item, n):
    if isinstance(item, (list, np.ndarray)):
        return np.asarray(item)

    return np.asarray([item for _ in range(n)])


def _fix_item(item):
    if isinstance(item, (list, np.ndarray)):
        return np.asarray(item).ravel()
    return item


def _get_length(item):
    """Return the length of item it item is a list, else 0."""
    if isinstance(item, (list, np.ndarray)):
        return np.asarray(item).size
    return 0


def _check_size(n_list):
    index = np.where(n_list > 0)
    n_list_reduced = n_list[index]
    first_item = n_list_reduced[0]
    if np.all(n_list_reduced == first_item):
        return first_item
    raise ValueError(
        "Each entry can either be a single item or a list of items. If an entry is a "
        "list, it's length must be equal to the length of other lists present in the "
        "system."
    )



# -*- coding: utf-8 -*-
"""
Solid-state NMR spectra simulation module for Python
====================================================

mrsimulator is an incredibly fast solid-state NMR spectra simulation package,
capable of simulating spectra from both crystalline and amorphous-like materials.


It aims to provide simple and efficient solutions to the solid-state NMR spectrum
simulation problem. It includes tools for users to create their model spin systems,
simulate and compare the simulation with the measurement, and perform least-squares
minimization, using a collection of pre-defined NMR methods.

See https://mrsimulator.readthedocs.io/en/stable/ for complete documentation.
"""
# version has to be specified at the start.
__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__copyright__ = "Copyright 2019-2021, The mrsimulator Project."
__credits__ = ["Deepansh J. Srivastava"]
__license__ = "BSD License"
__maintainer__ = "Deepansh J. Srivastava"
__status__ = "Beta"
__version__ = "0.6.0rc5"

import os

os.environ["OMP_NUM_THREADS"] = "1"
os.environ["OPENBLAS_NUM_THREADS"] = "1"
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["VECLIB_MAXIMUM_THREADS"] = "1"
# os.environ["NUMEXPR_NUM_THREADS"] = "1"

from .spin_system import Site  # lgtm [py/import-own-module] # noqa:F401
from .spin_system import Coupling  # lgtm [py/import-own-module]  # noqa:F401
from .spin_system import SpinSystem  # lgtm [py/import-own-module] # noqa:F401
from .simulator import Simulator
from .method.event import Event  # lgtm [py/import-own-module] # noqa:F401
from .method.spectral_dimension import (  # lgtm [py/import-own-module] # noqa:F401
    SpectralDimension,
)
from .method import Method  # lgtm [py/import-own-module] # noqa:F401
from mrsimulator.utils.importer import import_json
from mrsimulator.signal_processing import SignalProcessor
import json
from lmfit import Parameters


def save(
    filename: str,
    simulator: Simulator,
    signal_processors: list = None,
    params: Parameters = None,
    with_units: bool = True,
):
    """Serialize the Simulator, list of SignalProcessor, and lmfit Parameters objects
    to a .mrsim file.

    Args:
        str filename: The data is serialized to this file.
        sim: Simulator object.
        signal_processors: A list of PostSimulator objects corresponding to the methods
            in the Simulator object. Default is None.
        params: lmfit Parameters object. Default is None.
        bool with_units: If true, physical quantities are represented as string with
            units. The default is True.
    """
    py_dict = dict(simulator, signal_processors, params, with_units)

    with open(filename, "w", encoding="utf8") as outfile:
        json.dump(
            py_dict,
            outfile,
            ensure_ascii=False,
            sort_keys=False,
            allow_nan=False,
            separators=(",", ":"),
        )


def dict(
    simulator: Simulator,
    signal_processors: list = None,
    params: Parameters = None,
    with_units: bool = True,
):
    """Export the Simulator, list of SignalProcessor, and lmfit Parameters objects
    to a python dictionary.

    Args:
        sim: Simulator object.
        signal_processors: A list of PostSimulator objects corresponding to the methods
            in the Simulator object. Default is None.
        params: lmfit Parameters object. Default is None.
        bool with_units: If true, physical quantities are represented as string with
            units. The default is True.

    Return:
        Python dictionary
    """
    py_dict = simulator.json(True, True) if with_units else simulator.reduced_dict()
    if signal_processors is not None:
        py_dict["signal_processors"] = [item.json() for item in signal_processors]

    py_dict["params"] = None if params is None else params.dumps()

    return py_dict


def load(filename: str, parse_units: bool = True):
    """Load Simulator, list of SignalProcessor and optionally lmfit Parameters objects
    from the .mrsim file.

    Args:
        str filename: The location to the .mrsim file.
        bool parse_units: If true, parse the dictionary for units. The default is True.

    Return:
        Ordered List: Simulator, List[SignalProcessor], Parameters.
    """
    val = import_json(filename)
    return parse(val, parse_units)


def parse(py_dict, parse_units: bool = True):
    """Parse the dictionary object to respective Simulator, SignalProcessor and
    optionally lmfit Parameters object.

    Args:
        dict py_dict: Python dictionary representation of mrsimulator.
        bool parse_units: If true, parse the dictionary for units. Default is True.

    Return:
        Ordered List: Simulator, List[SignalProcessor], Parameters.
    """
    sim = Simulator.parse(py_dict, parse_units)

    signal_processors = (
        [
            SignalProcessor.parse_dict_with_units(item)
            for item in py_dict["signal_processors"]
        ]
        if "signal_processors" in py_dict
        else [SignalProcessor() for _ in sim.methods]
    )

    params = None
    if "params" in py_dict:
        val = py_dict["params"]
        params = None if val is None else Parameters().loads(s=val)

    return sim, signal_processors, params



# -*- coding: utf-8 -*-
"""
Function list:
    - Cosy
    - Inadequate
"""
from .base import BaseNamedMethod2D

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


class Cosy(BaseNamedMethod2D):
    r"""Simulate an infinite spinning COrrelation SpectroscopY spectrum.

    Args:
        channels: A list of isotope symbols over which the method will be applied.
        spectral_dimensions: A list of python dict. Each dict is contains keywords that
            describe the coordinates along a spectral dimension. The keywords along with
            its definition are:

            - count:
                An optional integer with the number of points, :math:`N`, along the
                dimension. The default value is 1024.
            - spectral_width:
                An `optional` float with the spectral width, :math:`\Delta x`, along the
                dimension in units of Hz. The default is 25 kHz.
            - reference_offset:
                An `optional` float with the reference offset, :math:`x_0` along the
                dimension in units of Hz. The default value is 0 Hz.
            - origin_offset:
                An `optional` float with the origin offset (Larmor frequency) along the
                dimension in units of Hz. The default value is None.
        magetic_flux_density: An `optional` float containing the macroscopic magnetic
            flux density, :math:`H_0`, of the applied external magnetic field in units
            of T. The default value is ``9.4``.
        rotor_angle: An `optional` float containing the angle between the sample
            rotation axis and the applied external magnetic field, :math:`\theta`, in
            units of rad. The default value is 0.9553166, i.e. the magic angle.
    note:
        The `rotor_frequency` parameter is fixed for this method.

    Return:
        A :class:`~mrsimulator.Method` instance.

    Example:
        >>> cosy = Cosy(
        ...     channels=["13C"],
        ...     magnetic_flux_density=7,  # in T
        ...     spectral_dimensions=[
        ...         {
        ...             "count": 256,
        ...             "spectral_width": 4.4e3,  # in Hz
        ...             "reference_offset": 2.2e3,  # in Hz
        ...         },
        ...         {
        ...             "count": 256,
        ...             "spectral_width": 4.4e3,  # in Hz
        ...             "reference_offset": 2.2e3,  # in Hz
        ...         },
        ...     ],
        ... )

    For two coupled spin-1/2 system, Cosy method selects 16 transition pathways.

        >>> sys = SpinSystem(
        ...     sites=[
        ...         Site(isotope='13C', isotropic_chemical_shift=10.0),
        ...         Site(isotope='13C', isotropic_chemical_shift=50.0,)
        ...     ],
        ...     couplings=[Coupling(site_index=[0, 1], isotropic_j=200)]
        ... )
        >>> pprint(cosy.get_transition_pathways(sys))
        [|-0.5, -0.5⟩⟨-0.5, 0.5| ⟶ |-0.5, -0.5⟩⟨-0.5, 0.5|,
         |-0.5, -0.5⟩⟨-0.5, 0.5| ⟶ |-0.5, -0.5⟩⟨0.5, -0.5|,
         |-0.5, -0.5⟩⟨-0.5, 0.5| ⟶ |-0.5, 0.5⟩⟨0.5, 0.5|,
         |-0.5, -0.5⟩⟨-0.5, 0.5| ⟶ |0.5, -0.5⟩⟨0.5, 0.5|,
         |-0.5, -0.5⟩⟨0.5, -0.5| ⟶ |-0.5, -0.5⟩⟨-0.5, 0.5|,
         |-0.5, -0.5⟩⟨0.5, -0.5| ⟶ |-0.5, -0.5⟩⟨0.5, -0.5|,
         |-0.5, -0.5⟩⟨0.5, -0.5| ⟶ |-0.5, 0.5⟩⟨0.5, 0.5|,
         |-0.5, -0.5⟩⟨0.5, -0.5| ⟶ |0.5, -0.5⟩⟨0.5, 0.5|,
         |-0.5, 0.5⟩⟨0.5, 0.5| ⟶ |-0.5, -0.5⟩⟨-0.5, 0.5|,
         |-0.5, 0.5⟩⟨0.5, 0.5| ⟶ |-0.5, -0.5⟩⟨0.5, -0.5|,
         |-0.5, 0.5⟩⟨0.5, 0.5| ⟶ |-0.5, 0.5⟩⟨0.5, 0.5|,
         |-0.5, 0.5⟩⟨0.5, 0.5| ⟶ |0.5, -0.5⟩⟨0.5, 0.5|,
         |0.5, -0.5⟩⟨0.5, 0.5| ⟶ |-0.5, -0.5⟩⟨-0.5, 0.5|,
         |0.5, -0.5⟩⟨0.5, 0.5| ⟶ |-0.5, -0.5⟩⟨0.5, -0.5|,
         |0.5, -0.5⟩⟨0.5, 0.5| ⟶ |-0.5, 0.5⟩⟨0.5, 0.5|,
         |0.5, -0.5⟩⟨0.5, 0.5| ⟶ |0.5, -0.5⟩⟨0.5, 0.5|]
    """

    def __new__(cls, **kwargs):
        return cls.update(super().__new__(cls, **kwargs))

    @staticmethod
    def update(method):
        # setting transition symmetry elements
        sd = method.spectral_dimensions
        sd[0].events[0].transition_query.P = {"channel-1": [[-1]]}
        sd[1].events[0].transition_query.P = {"channel-1": [[-1]]}

        # method description
        method.description = (
            "Simulate an infinite spinning COrrelation SpectroscopY spectrum."
        )
        return method


class Inadequate(BaseNamedMethod2D):
    r"""Simulate an infinite spinning Incredible Natural Abundance DoublE QUAntum
    Transfer Experiment spectrum.

    Args:
        channels: A list of isotope symbols over which the method will be applied.
        spectral_dimensions: A list of python dict. Each dict is contains keywords that
            describe the coordinates along a spectral dimension. The keywords along with
            its definition are:

            - count:
                An optional integer with the number of points, :math:`N`, along the
                dimension. The default value is 1024.
            - spectral_width:
                An `optional` float with the spectral width, :math:`\Delta x`, along the
                dimension in units of Hz. The default is 25 kHz.
            - reference_offset:
                An `optional` float with the reference offset, :math:`x_0` along the
                dimension in units of Hz. The default value is 0 Hz.
            - origin_offset:
                An `optional` float with the origin offset (Larmor frequency) along the
                dimension in units of Hz. The default value is None.
        magetic_flux_density: An `optional` float containing the macroscopic magnetic
            flux density, :math:`H_0`, of the applied external magnetic field in units
            of T. The default value is ``9.4``.
        rotor_angle: An `optional` float containing the angle between the sample
            rotation axis and the applied external magnetic field, :math:`\theta`, in
            units of rad. The default value is 0.9553166, i.e. the magic angle.
    note:
        The `rotor_frequency` parameter is fixed for this method.

    Return:
        A :class:`~mrsimulator.Method` instance.

    Example:
        >>> inadequate = Inadequate(
        ...     channels=["13C"],
        ...     magnetic_flux_density=7,  # in T
        ...     spectral_dimensions=[
        ...         {
        ...             "count": 256,
        ...             "spectral_width": 4.4e3,  # in Hz
        ...             "reference_offset": 2.2e3,  # in Hz
        ...         },
        ...         {
        ...             "count": 256,
        ...             "spectral_width": 4.4e3,  # in Hz
        ...             "reference_offset": 2.2e3,  # in Hz
        ...         },
        ...     ],
        ... )

    For two coupled spin-1/2 system, Cosy method selects 16 transition pathways.

        >>> sys = SpinSystem(
        ...     sites=[
        ...         Site(isotope='13C', isotropic_chemical_shift=10.0),
        ...         Site(isotope='13C', isotropic_chemical_shift=50.0,)
        ...     ],
        ...     couplings=[Coupling(site_index=[0, 1], isotropic_j=200)]
        ... )
        >>> pprint(inadequate.get_transition_pathways(sys))
        [|-0.5, -0.5⟩⟨0.5, 0.5| ⟶ |-0.5, -0.5⟩⟨-0.5, 0.5|,
         |-0.5, -0.5⟩⟨0.5, 0.5| ⟶ |-0.5, -0.5⟩⟨0.5, -0.5|,
         |-0.5, -0.5⟩⟨0.5, 0.5| ⟶ |-0.5, 0.5⟩⟨0.5, 0.5|,
         |-0.5, -0.5⟩⟨0.5, 0.5| ⟶ |0.5, -0.5⟩⟨0.5, 0.5|]
    """

    def __new__(cls, **kwargs):
        return cls.update(super().__new__(cls, **kwargs))

    @staticmethod
    def update(method):
        # setting transition symmetry elements
        sd = method.spectral_dimensions
        sd[0].events[0].transition_query.P = {"channel-1": [[-1, -1]]}
        sd[1].events[0].transition_query.P = {"channel-1": [[-1]]}

        # method description
        method.description = (
            "Simulate an infinite spinning Incredible Natural Abundance DoublE QUAntum "
            "Transfer Experiment spectrum."
        )
        return method



# -*- coding: utf-8 -*-
import mrsimulator.signal_processing as sp
import numpy as np
from lmfit import Parameters
from mrsimulator import Simulator
from mrsimulator.method import NAMED_METHODS

__author__ = ["Maxwell C Venetos", "Deepansh Srivastava"]
__email__ = ["maxvenetos@gmail.com", "srivastava.89@osu.edu"]

START = "sys_"
ENCODING_PAIRS = [
    ["spin_systems[", START],
    ["].abundance", "_abundance"],
    # site
    ["].sites[", "_site_"],
    ["].isotropic_chemical_shift", "_isotropic_chemical_shift"],
    ["].shielding_symmetric.", "_shielding_symmetric_"],
    ["].quadrupolar.", "_quadrupolar_"],
    # coupling
    ["].couplings[", "_coupling_"],
    ["].isotropic_j", "_isotropic_j"],
    ["].j_symmetric.", "_j_symmetric_"],
    ["].dipolar.", "_dipolar_"],
]

DECODING_PAIRS = [
    ["spin_systems.", "sys_"],
    [".abundance", "_abundance"],
    # site
    [".sites.", "_site_"],
    [".isotropic_chemical_shift", "_isotropic_chemical_shift"],
    [".shielding_symmetric.", "_shielding_symmetric_"],
    [".quadrupolar.", "_quadrupolar_"],
    # coupling
    [".couplings.", "_coupling_"],
    [".isotropic_j", "_isotropic_j"],
    [".j_symmetric.", "_j_symmetric_"],
    [".dipolar.", "_dipolar_"],
]

EXCLUDE = [
    "property_units",
    "isotope",
    "name",
    "label",
    "description",
    "transition_pathways",
]

POST_SIM_DICT = {
    "Gaussian": {"FWHM": "FWHM"},
    "Exponential": {"FWHM": "FWHM"},
    "Scale": {"factor": "factor"},
    "ConstantOffset": {"offset": "offset"},
    "Linear": {"amplitude": "amplitude", "offset": "offset"},
}


def _str_encode(my_string):
    """LMFIT Parameters class does not allow for names to include special characters.
    This function replaces '[', ']', and '.' to '_' to comply with LMFIT rules.

    Args:
        my_string: A string object

    Returns:
        String object.
    """
    for item in ENCODING_PAIRS:
        my_string = my_string.replace(*item)
    return my_string


def _str_decode(my_string):
    """Parse the string for objects and indexes and return a list.

    Args:
        my_string: A string object

    Returns:
        List of strings with strings resresenting mrsimulator objects and indexes.

    Example:
        >>> string = 'sys_0_site_0_isotropic_chemical_shift'
        >>> _str_decode(string)
        ['spin_systems', '0', 'sites', '0', 'isotropic_chemical_shift']

        >>> string = 'sys_10_abundance'
        >>> _str_decode(string)
        ['spin_systems', '10', 'abundance']
    """
    for item in DECODING_PAIRS:
        my_string = my_string.replace(*item[::-1])
    my_string = my_string.split(".")
    return my_string


def _list_of_dictionaries(my_list):
    """Helper function for traverse_dictionaries function which will return a list of
    dictionaries.

    Args:
        my_list: A list object

    Returns:
        List Object.
    """
    return [item.dict() for item in my_list]


def _traverse_dictionaries(instance, parent="spin_systems"):
    """Parses through the instance object contained within the parent object and return
    a list of attributes that are populated.

    Args:
        instance: An instance object from the parent object.
        parent: a string object used to create the addresses of the SpinSystem
            attributes.

    Returns:
        List Object.
    """
    if isinstance(instance, list):
        return [
            value
            for i, obj in enumerate(instance)
            for value in _traverse_dictionaries(obj, _str_encode(f"{parent}[{i}]"))
        ]

    if isinstance(instance, dict):
        return [
            item
            for key, value in instance.items()
            if key not in EXCLUDE and value is not None
            for item in (
                _traverse_dictionaries(value, _str_encode(f"{parent}.{key}"))
                if isinstance(value, (dict, list))
                else [_str_encode(f"{parent}.{key}")]
            )
        ]

    return []


def _post_sim_LMFIT_params(params, process, index):
    """Creates a LMFIT Parameters object for SignalProcessor operations involved in
    spectrum fitting.

    Args:
        params: LMFIT parameters object.
        process: SignalProcessor object at index *index*.
        int index: List index of the SingalProcessor object.

    Returns:
        Parameters object.
    """
    _ = [
        params.add(
            name=f"SP_{index}_operation_{i}_{operation.__class__.__name__}_{attr}",
            value=operation.__getattribute__(attr),
        )
        for i, operation in enumerate(process.operations)
        if operation.__class__.__name__ in POST_SIM_DICT
        for attr in POST_SIM_DICT[operation.__class__.__name__]
    ]


def make_signal_processor_params(processors: list):
    """Parse the list of SignalProcessor objects for a list of LMFIT parameters.

    Args:
        processors: List of SignalProcessor objects. The order of the list mush match
            the order of the methods in the Simulator object.
    """
    processors = processors if isinstance(processors, list) else [processors]

    correct_type = all([isinstance(obj, sp.SignalProcessor) for obj in processors])
    if not correct_type:
        raise ValueError("Expecting a list of `SignalProcessor` objects.")

    params = Parameters()
    _ = [_post_sim_LMFIT_params(params, obj, i) for i, obj in enumerate(processors)]
    return params


def _get_simulator_object_value(sim, string):
    """Parse the string representing the Simulator object dictionary tree format, and
    return its value.

    Args:
        sim: The simulator object.
        string: A string representing the Simulator object dictionary tree format.

    Returns:
        Float. The value of the object.

    Example:
        >>> site = Site(isotropic_chemical_shift=-431)
        >>> sys = SpinSystem(sites=[site], abundance=23)
        >>> sim = Simulator()
        >>> sim.spin_systems.append(sys)

        >>> string = 'sys_0_site_0_isotropic_chemical_shift'
        >>> _get_simulator_object_value(sim, string)
        -431.0

        >>> string = 'sys_0_abundance'
        >>> _get_simulator_object_value(sim, string)
        23.0
    """
    string = _str_decode(string)
    obj = sim
    for attr in string:
        obj = obj[int(attr)] if attr.isnumeric() else obj.__getattribute__(attr)
    return obj


def make_simulator_params(sim: Simulator, include={}):
    """Parse the Simulator object for a list of LMFIT parameters.

    Args:
        Simulator sim: Simulator object.
    """
    params = Parameters()
    if not isinstance(sim, Simulator):
        raise ValueError(f"Expecting a `Simulator` object, found {type(sim).__name__}.")

    temp_list = _traverse_dictionaries(_list_of_dictionaries(sim.spin_systems))

    # get total abundance scaling factor
    length = len(sim.spin_systems)
    abundance_scale = 100 / sum([sys.abundance for sys in sim.spin_systems])

    # expression for the last abundance.
    last_abundance = f"{length - 1}_abundance"
    expression = "-".join([f"{START}{i}_abundance" for i in range(length - 1)])
    expression = "100" if expression == "" else f"100-{expression}"

    for items in temp_list:
        value = _get_simulator_object_value(sim, items)
        if "_eta" in items:
            params.add(name=items, value=value, min=0, max=1)

        # last_abundance should come before abundance
        elif last_abundance in items:
            params.add(name=items, value=value, min=0, max=100, expr=expression)

        elif "abundance" in items:
            params.add(name=items, value=value * abundance_scale, min=0, max=100)
        else:
            params.add(name=items, value=value)

    get_simulator_method_parameters(params, sim, include)
    return params


def get_simulator_method_parameters(params: Parameters, sim: Simulator, include={}):
    """LMFIT parameters for the method attribute in `include`

    Args:
        Parameters params: LMFIT parameters object.
        Simulator sim: Simulator object.
        set include: Set of methtod attributes to include.
    """
    if "rotor_frequency" in include:
        values = [
            [
                ev.rotor_frequency
                for sp in mth.spectral_dimensions
                for ev in sp.events
                if ev.rotor_frequency != 1e12
            ]
            for i, mth in enumerate(sim.methods)
            if mth.name in NAMED_METHODS
        ]

        _ = [
            params.add(
                name=f"mth_{i}_rotor_frequency",
                value=val[0],
                min=val[0] - 100,
                max=val[0] + 100,
            )
            for i, val in enumerate(values)
            if val != []
        ]

    return params


def make_LMFIT_parameters(sim: Simulator, processors: list = None, include={}):
    """An alias of `make_LMFIT_params` function."""
    return make_LMFIT_params(sim, processors)


def make_LMFIT_params(sim: Simulator, processors: list = None, include={}):
    r"""Parse the Simulator and PostSimulator objects for a list of LMFIT parameters.

    Args:
        Simulator sim: Simulator object.
        list processors: List of SignalProcessor objects. The order must match the order
            of methods within the simulator object.
        set include: set of keywords from the method object to include as a fitting
            parameter. Default is {}.

    The parameter name assocciated with the spin system within Simulator object is
    generated using the following nomenclature- *sys_i_site_j_attribute1_attribute2*
    for attribute with signature ``sim.spin_systems[i].sites[j].attribute1.attribute2``

    Here, *sys_i* refers to the spin system at index *i*, *site_j* refers to the site at
    index *j* with in the :math:i^\text{th} spin system, and *attribute1* and
    *attribute2* are the site attrbiutes.

    **For examples:**

    ``sim.spin_systems[1].sites[0].isotropic_chemical_shift`` parametrizes to
    *sys_1_site_0_isotropic_chemical_shift* while
    ``sim.spin_systems[0].sites[1].quadrupolar.Cq`` to *sys_0_site_1_quadrupolar_Cq*.

    Returns:
        LMFIT Parameters object.
    """
    params = Parameters()
    params.update(make_simulator_params(sim, include))

    proc = make_signal_processor_params(processors) if processors is not None else None
    params.update(proc) if proc is not None else None

    return params


def _update_simulator_from_LMFIT_params(params, sim: Simulator):
    """Parse the string representing the Simulator object dictionary tree format, and
    set its value to the input.

    Args:
        sim: The simulator object.
        string: A string representing the Simulator object dictionary tree format.
        value: The value to assign.

    Example:
        if params['sys_i_site_j_isotropic_chemical_shift'].value = 120
        the function sets, sim.spin_systems[i].sites[j].isotropic_chemical_shift = 120.0
    """

    def set_sys_value(obj, key, value):
        ids = _str_decode(key)
        for attr in ids[:-1]:
            obj = obj[int(attr)] if attr.isnumeric() else obj.__dict__[attr]
        obj.__dict__[ids[-1]] = value

    def set_mth_value(obj, key, value):
        index = int(key.split("_")[1])
        _ = [
            sp.events[0].__setattr__("rotor_frequency", value)
            for sp in obj.__dict__["methods"][index].spectral_dimensions
            if sp.events[0].rotor_frequency != 1e12
        ]

    values = params.valuesdict()
    _ = [set_sys_value(sim, k, v) for k, v in values.items() if "sys_" in k]
    _ = [set_mth_value(sim, k, v) for k, v in values.items() if "mth_" in k]


def _update_processors_from_LMFIT_params(params, processors: list):
    """Updates SignalProcessor operation arguments from a LMFIT Parameters object.

    Args:
        params: LMFIT Parameters object.
        processors: list of SignalProcessor objects.

    Example:
        if params['SP_i_operations_j_Scale_factor'].value = 10
        the function sets, processors[i].operations[j].Scale.factor = 10.0
    """

    def set_value(obj, key, value):
        ids = key.split("_")
        # The signal processor index, operation indexes, and function argument
        sp, op, arg = int(ids[1]), int(ids[3]), POST_SIM_DICT[ids[4]][ids[5]]
        obj[sp].__dict__["operations"][op].__dict__[arg] = value

    values = params.valuesdict()
    _ = [set_value(processors, k, v) for k, v in values.items() if "operation_" in k]


def update_mrsim_obj_from_params(params, sim: Simulator, processors: list = None):
    """Update the mrsimulator Simulator and SignalProcessor objects from the lmfit
    Parameters obj

    Args:
        params: Parameters object containing parameters for OLS minimization.
        sim: Simulator object.
        processors: A list of SignalProcessor objects corresponding to the methods in
            the Simulator object.
    """
    _update_simulator_from_LMFIT_params(params, sim)
    _update_processors_from_LMFIT_params(params, processors)


def get_correct_data_order(data):
    """If data has negative increment, reverse the data."""
    y = data.y[0].components[0]
    index = [-i - 1 for i, x in enumerate(data.x) if x.increment.value < 0]
    return y if index == [] else np.flip(y, axis=tuple(index))


def LMFIT_min_function(
    params: Parameters, sim: Simulator, processors: list = None, sigma: list = None
):
    """The simulation routine to calculate the vector difference between simulation and
    experiment based on the parameters update.

    Args:
        params: Parameters object containing parameters for OLS minimization.
        sim: Simulator object.
        processors: A list of PostSimulator objects corresponding to the methods in the
            Simulator object.
        sigma: A list of standard deviations corresponding to the experiments in the
            Simulator.methods attribute
    Returns:
        Array of the differences between the simulation and the experimental data.
    """
    processors = processors if isinstance(processors, list) else [processors]
    sigma = [1.0 for _ in sim.methods] if sigma is None else sigma
    sigma = sigma if isinstance(sigma, list) else [sigma]

    update_mrsim_obj_from_params(params, sim, processors)

    sim.run()

    processed_data = [
        item.apply_operations(data=data.simulation)
        for item, data in zip(processors, sim.methods)
    ]

    diff = np.asarray([])
    for processed_datum, mth, sigma_ in zip(processed_data, sim.methods, sigma):
        datum = 0
        for decomposed_datum in processed_datum.y:
            datum += decomposed_datum.components[0].real

        exp_data = get_correct_data_order(mth.experiment)
        diff = np.append(diff, (exp_data - datum) / sigma_)
    return diff


def bestfit(sim: Simulator, processors: list = None):
    """Return a list of best fit spectrum ordered relative to the methods in the
    simulator object.

    Args:
        Simulator sim: The simulator object.
        list processors: List of SignalProcessor objects ordered according to the
            methods in the simulator object.
    """
    processors = processors if isinstance(processors, list) else [processors]
    sim.run()

    return [
        proc.apply_operations(data=mth.simulation).real
        for mth, proc in zip(sim.methods, processors)
    ]


def add_csdm_dvs(data):
    new_data = data.split()
    new_csdm = 0
    for item in new_data:
        new_csdm += item
    return new_csdm if new_data != [] else None


def residuals(sim: Simulator, processors: list = None):
    """Return a list of residuals corresponsing to the best fit spectrum. The list is
    based on the order of methods in the simulator object.

    Args:
        Simulator sim: The simulator object.
        list processors: List of SignalProcessor objects ordered according to the
            methods in the simulator object."""
    fits = bestfit(sim, processors)
    residual_ = [add_csdm_dvs(item) for item in fits]

    for res, mth in zip(residual_, sim.methods):
        exp_data = get_correct_data_order(mth.experiment)
        res.y[0].components[0] -= exp_data
        res.y[0].components[0] *= -1

    return residual_



# -*- coding: utf-8 -*-
import numpy as np
from mrsimulator.utils.abstract_list import AbstractList

from . import Transition

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


class TransitionList(AbstractList):
    def __init__(self, data: list = []):
        super().__init__([self._check_for_transition_object(item) for item in data])

    @staticmethod
    def _check_for_transition_object(item):
        if isinstance(item, dict):
            return Transition(**item)
        if not isinstance(item, Transition):
            raise ValueError(
                "Expecting a Transition object or an equivalent python dict object, "
                f"instead found {item.__class__.__name__}."
            )
        return item

    def __setitem__(self, index, item):
        self._list[index] = self._check_for_transition_object(item)

    def append(self, item):
        super().append(self._check_for_transition_object(item))

    def filter(self, P=None, PP=None, D=None):
        """Filter a list of transitions to satisfy the filtering criterion.
        Args:
            list P: A list of `N` (m_final - m_initial) values, where `N` is the
                total number of sites within the spin system.
            list D: A list of `N` (m_final^2 - m_initial^2) values, where `N` is the
                total number of sites within the spin system.
        """

        # to think
        # - filter based on transition.
        # - filter based on state.

        if P is PP is D is None:
            return self

        ts = self._list.copy()

        if P is not None:
            ts = TransitionList([item for item in ts if np.allclose(item.P, P)])

        # if PP is not None:
        #     ts = TransitionList([item for item in ts if np.allclose(item.PP, PP)])

        if D is not None:
            ts = TransitionList([item for item in ts if np.allclose(item.D, D)])
        # if transitions is not None:
        #     for transition in transitions:
        #         ts = TransitionList(
        #             [item for item in ts if item == Transition(**transition)]
        #         )
        # if start_state is not None:
        #     ts = [item for item in ts if ts.initial == start_state]
        return ts

        # lst = self._list
        # for i, element in enumerate(search):
        #     if isinstance(element, int):
        #         lst = [item for item in lst if item.delta_m(i) == element]
        #     if isinstance(element, list):
        #         lst = [
        #             item
        #             for item in lst
        #             if (item.initial[i] == element[1] and item.final[i] == element[0])
        #         ]
        # return lst
        #     return TransitionList(
        #         [item for item in self._list if np.all(item.initial == delta_ms)]
        #     )

        # if delta_ms is not None:
        #     return TransitionList(
        #         [item for item in self._list if np.all(item.delta_ms == delta_ms)]
        #     )
        # return TransitionList([item for item in self._list if item.p == p])


class TransitionPathway(TransitionList):
    """Base TransitionPathway class is a list of connected Transitions.

    Example:
        >>> from mrsimulator.transition import TransitionPathway, Transition
        >>> t1 = Transition(initial = [0.5, 0.5], final = [0.5, -0.5])
        >>> t2 = Transition(initial=[0.5, 0.5], final=[-0.5, 0.5])
        >>> path = TransitionPathway([t1, t2])
        >>> path
        |0.5, -0.5⟩⟨0.5, 0.5| ⟶ |-0.5, 0.5⟩⟨0.5, 0.5|
    """

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return " ⟶ ".join([repr(item) for item in self._list])

    def json(self) -> dict:
        """Parse the class object to a JSON compliant python dictionary object.

        Example:
            >>> pprint(path.json())
            [{'final': [0.5, -0.5], 'initial': [0.5, 0.5]},
             {'final': [-0.5, 0.5], 'initial': [0.5, 0.5]}]
        """
        return [item.json() for item in self._list]

    def tolist(self):
        """Expand TransitionPathway to a Python list.

        Example:
            >>> path.tolist()
            [0.5, 0.5, 0.5, -0.5, 0.5, 0.5, -0.5, 0.5]
        """
        return list(np.asarray([item.tolist() for item in self._list]).ravel())



# -*- coding: utf-8 -*-
from typing import ClassVar

import numpy as np
from mrsimulator.method import Method

from .utils import generate_method_from_template
from .utils import METHODS_DATA

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"

generic_args = r"""

Parameters
----------

name: str (optional).
    The value is the name or id of the method. The default value is None.

label: str (optional).
    The value is a label for the method. The default value is None.

description: str (optional).
    The value is a description of the method. The default value is None.

experiment: CSDM or ndarray (optional).
    An object holding the experimental measurement for the given method, if
    available. The default value is None.

channels: list (optional).
    The value is a list of isotope symbols over which the given method applies.
    An isotope symbol is given as a string with the atomic number followed by its
    atomic symbol, for example, '1H', '13C', and '33S'. The default is an empty
    list.
    The number of isotopes in a `channel` depends on the method. For example, a
    `BlochDecaySpectrum` method is a single channel method, in which case, the
    value of this attribute is a list with a single isotope symbol, ['13C'].

spectral_dimensions: List of :ref:`spectral_dim_api` or dict objects (optional).
    The number of spectral dimensions depends on the given method. For example, a
    `BlochDecaySpectrum` method is a one-dimensional method and thus requires a
    single spectral dimension. The default is a single default
    :ref:`spectral_dim_api` object.

magetic_flux_density: float (optional)
    A global value for the macroscopic magnetic flux density, :math:`H_0`,
    of the applied external magnetic field in units of T. The default is ``9.4``.

rotor_angle: float (optional)
    A global value for the angle between the sample rotation axis and the
    applied external magnetic field, :math:`\theta`, in units of rad. The default
    value is ``0.9553166``, i.e. the magic angle.
"""

args_freq = r"""
rotor_frequency: float (optional)
    A global value for the sample spinning frequency, :math:`\nu_r`, in units of Hz.
    The default value is ``0``.
"""

args_affine = r"""
affine_matrix: np.ndarray or list (optional)
    An affine transformation square matrix,
    :math:`\mathbf{A} \in \mathbb{R}^{n \times n}`, where `n` is the number of
    spectral dimensions. The affine operation follows
    :math:`\mathbf{V}^\prime = \mathbf{A} \cdot \mathbf{V}`,
    where :math:`\mathbf{V}\in\mathbb{R}^n` and :math:`\mathbf{V}^\prime\in\mathbb{R}^n`
    are the initial and transformed frequency coordinates.
"""

# additional_args = args_freq + args_affine

returns = r"""
Return
------
    A :class:`~mrsimulator.Method` instance.
"""

notes = r"""
Note
----

If any parameter is defined outside of the `spectral_dimensions` list, the value
of those parameters is considered global. In a multi-event method, you may also
assign parameter values to individual events.
"""

docstring_generic = "".join([generic_args, args_freq, returns, notes])
docstring_1D = "".join([generic_args, args_freq, returns])

# BlochDecaySpectrum
BlochDecaySpectrum = generate_method_from_template(
    METHODS_DATA["Bloch_decay"], docstring_1D
)

BlochDecayCentralTransitionSpectrum = generate_method_from_template(
    METHODS_DATA["Bloch_decay_central_transition"], docstring_1D
)
BlochDecayCTSpectrum = BlochDecayCentralTransitionSpectrum

# generic 1D method
Method1D = generate_method_from_template(METHODS_DATA["Method1D"], docstring_generic)

# generic 2D method
docstring_2D = "".join([generic_args, args_affine, returns, notes])
Method2D = generate_method_from_template(METHODS_DATA["Method2D"], docstring_2D)


class BaseNamedMethod(Method):
    def __new__(cls, **kwargs):
        n_dim = cls.ndim
        spectral_dimensions = cls.check_for_spectral_dimensions(kwargs, n_dim)
        cls.check_for_transition_query(spectral_dimensions)
        cls.check_for_events(spectral_dimensions)

        if "name" not in kwargs:
            kwargs["name"] = cls.__name__

        method_fn = Method1D if n_dim == 1 else Method2D

        return method_fn(spectral_dimensions, **kwargs)

    @staticmethod
    def check_for_spectral_dimensions(py_dict, n=1):
        """If spectral_dimensions in py_dict, extract and then remove from py_dict."""

        if "spectral_dimensions" not in py_dict:
            return [{}] * n

        spectral_dimensions = py_dict["spectral_dimensions"]
        m = len(spectral_dimensions)
        if m == n:
            py_dict.pop("spectral_dimensions")
            return spectral_dimensions
        raise ValueError(f"Method requires exactly {n} spectral dimensions, given {m}.")

    @classmethod
    def message(cls, attr):
        return f"`{attr}` value cannot be modified for {cls.__name__} method."

    @classmethod
    def check_for_transition_query(cls, spectral_dimensions=[{}, {}]):
        """Check for transition_query attribute inside the spectral_dimensions.
        The transition query are not allowed for NamedMethods."""
        if isinstance(spectral_dimensions[0], dict):
            check = [
                "transition_query" in event.keys()
                for item in spectral_dimensions
                if "events" in item.keys()
                for event in item["events"]
            ]

            if np.any(check):
                raise ValueError(cls.message("transition_query"))

    @classmethod
    def check_for_events(cls, spectral_dimensions=[{}, {}]):
        """Check for events attribute inside the spectral_dimensions. Events are not
        allowed for NamedMethods."""
        if isinstance(spectral_dimensions[0], dict):
            check = ["events" in item.keys() for item in spectral_dimensions]

            if np.any(check):
                raise ValueError(cls.message("events"))


class BaseNamedMethod2D(BaseNamedMethod):
    ndim: ClassVar[int] = 2



# -*- coding: utf-8 -*-
"""The Event class."""
from typing import ClassVar
from typing import Dict
from typing import Union

import numpy as np
from pydantic import validator

from ._base import ModuleOperation
from .utils import _get_broadcast_shape
from .utils import _str_to_quantity
from .utils import CONST

__author__ = "Deepansh Srivastava"
__email__ = "srivastava.89@osu.edu"


class AffineTransformation(ModuleOperation):
    dim_index: int = 0
    dv_index: Union[int, list, tuple] = None  # if none apply to all
    module_name: ClassVar = __name__

    @property
    def function(self):
        return "affine"


class Shear(AffineTransformation):
    r"""Apply a shear parallel to dimension at index parallel and normal to dimension
    at index dim_index.

    The shear function is an apodization with the following form

    .. math::
        f(x) = e^{-i 2\pi x_0 x_1 a_0},

    where :math:`x_0` are the coordinates of the dimension at index `parallel`,
    :math:`x_1` are the coordinates of the dimension at index `dim_index`, and
    :math:`a_0` is the shear constant.

    Args:
        str factor: The shear factor is given as a string with a value and a unit. The
            default value is 0.
        int dim_index: The shear is applied normal to the CSDM dimension at this index.
            The default is the dimension at index 0.
        int parallel: The shear is applied parallel to the CSDM dimension at this index.
            The default is the dimension at index 1.
        int dv_index: The index of the CSDM dependent variable where the operation is
            applied. If the value is None, the operation will be applied to every
            dependent variable.

    Example
    -------

    >>> operation = sp.affine.Shear(factor='143.4 Hz', dim_index=0, parallel=1)
    """

    factor: Union[float, str] = 0
    parallel: int = 1
    property_units: Dict = {"factor": CONST}

    @validator("factor")
    def str_to_quantity(cls, v, values):
        return _str_to_quantity(v, values, "factor")

    # class Config:
    #     validate_assignment = True

    def operate(self, data):
        """Applies the operation.

        Args:
            data: CSDM object.
        """
        dims = data.dimensions
        n_dim = len(dims)

        x, y = dims[self.dim_index], dims[self.parallel]
        x_value, y_value = [self.get_coordinates(_).value for _ in [x, y]]

        vector_x = _get_broadcast_shape(x_value, self.dim_index, n_dim)
        vector_y = _get_broadcast_shape(y_value, self.parallel, n_dim)

        xy = vector_x * vector_y
        unit = 1 / self.property_units["factor"]

        multiplier = unit.to(x.increment.unit * y.increment.unit).value

        # apodization by exp(-i 2π w t a), where w is freq, t is time, and
        # a is the factor.
        apodization_vector = np.exp(-2j * np.pi * xy * self.factor * multiplier)

        dv_indexes = self._get_dv_indexes(self.dv_index, n=len(data.y))

        for i in dv_indexes:
            datum = data.y[i].components
            datum *= apodization_vector

        return data


class Scale(AffineTransformation):
    r"""Scale the dimension along the specified dimension index.

    Args:
        str factor: The scaling factor. The default is 1.
        int dim_index: The index of the CSDM dimension to scale. The default is the
            dimension at index 0.

    Example
    -------

    >>> operation = sp.affine.Scale(factor=2.14, dim_index=0)
    """

    factor: Union[float, str] = 1
    property_units: Dict = {"factor": CONST}

    @validator("factor")
    def str_to_quantity(cls, v, values):
        return _str_to_quantity(v, values, "factor")

    def operate(self, data):
        """Applies the operation.

        Args:
            data: CSDM object.
        """
        data_ref = data.x[self.dim_index]
        data_ref.increment *= self.factor
        data_ref.coordinates_offset *= self.factor
        data_ref.reciprocal.coordinates_offset /= self.factor
        return data


# class Translate(AbstractAffineTransformation):
#     r"""Apodize a dependent variable of the CSDM object with a Gaussian function.

#     The apodization function follows

#     .. math::
#         f(x) = e^{-2 \pi^2 \sigma^2  x^2},

#     where :math:`x` are the coordinates of the dimension, and :math:`\sigma` is the
#     standard deviation. The relationship between the standard deviation,
#     :math:`\sigma`,
#     and the full width at half maximum of the reciprocal domain Gaussian function
#     follows

#     .. math::
#         \sigma = \frac{\text{FWHM}}{2\sqrt{2\ln 2}}.

#     Args:
#         str FWHM: The full width at half maximum, FWHM, of the reciprocal domain
#             Gaussian function, given as a string with a value and a unit. The default
#             value is 0.
#         int dim_index: The index of the CSDM dimension along which the operation is
#             applied. The default is the dimension at index 0.
#         int dv_index: The index of the CSDM dependent variable where the operation is
#             applied. If the value is None, the operation will be applied to every
#             dependent variable.

#     Example
#     -------

#     >>> operation4 = sp.apodization.Gaussian(FWHM='143.4 Hz', dim_index=0, dv_index=0)
#     """

#     factor: Union[float, str] = 1
#     property_units: Dict = {"factor": CONST}

#     @validator("factor")
#     def str_to_quantity(cls, v, values):
#         return _str_to_quantity(v, values, "factor")

#     # class Config:
#     #     validate_assignment = True

#     def operate(self, data):
#         """
#         Applies the operation for which the class is named for.

#         data: CSDM object
#         dep_var: int. The index of the dependent variable to apply operation to.
#         """
#         # data.x[self.dim_index].increment *= self.factor
#         data.x[self.dim_index].coordinates_offset += (
#             self.factor * self.property_units["factor"]
#         )
#         return data



# -*- coding: utf-8 -*-
import numpy as np
import pytest
from mrsimulator.method.transition_query import TransitionQuery
from mrsimulator.methods import FiveQ_VAS
from mrsimulator.methods import SevenQ_VAS
from mrsimulator.methods import ThreeQ_VAS

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"

methods = [ThreeQ_VAS, FiveQ_VAS, SevenQ_VAS]
names = ["ThreeQ_VAS", "FiveQ_VAS", "SevenQ_VAS"]


def test_MQ_VAS_rotor_freq():
    def error(name):
        return f"`rotor_frequency` value cannot be modified for {name} method."

    for name, method in zip(names, methods):
        e = error(name)
        with pytest.raises(ValueError, match=f".*{e}.*"):
            method(rotor_frequency=10, spectral_dimensions=[{}, {}])


def test_MQ_VAS_spectral_dimension_count():
    e = "Method requires exactly 2 spectral dimensions, given 1."
    for _, method in zip(names, methods):
        with pytest.raises(ValueError, match=f".*{e}.*"):
            method(spectral_dimensions=[{}])


def test_MQ_VAS_setting_transition_query():
    def error(name):
        return f"`transition_query` value cannot be modified for {name} method."

    for name, method in zip(names, methods):
        e = error(name)
        with pytest.raises(ValueError, match=f".*{e}.*"):
            method(
                spectral_dimensions=[
                    {"events": [{"transition_query": {"P": [-1]}}]},
                    {},
                ],
            )


def test_MQ_VAS_affine():
    sites = ["87Rb", "27Al", "51V"]
    spins = [1.5, 2.5, 3.5]
    k_MQ_MAS = {
        3: {1.5: 21 / 27, 2.5: 114 / 72, 3.5: 303 / 135, 4.5: 546 / 216},
        5: {2.5: 150 / 72, 3.5: 165 / 135, 4.5: 570 / 216},
        7: {3.5: 483 / 135, 4.5: 84 / 216},
        9: {4.5: 1116 / 216},
    }
    for j, method in enumerate(methods):
        for i, isotope in zip(spins[j:], sites[j:]):
            meth = method(channels=[isotope])
            k = k_MQ_MAS[3 + 2 * j][i]
            assert meth.spectral_dimensions[0].events[0].fraction == 1
            assert meth.spectral_dimensions[1].events[0].fraction == 1
            assert np.allclose(meth.affine_matrix, [1 / (k + 1), k / (k + 1), 0, 1])


def test_3Q_VAS_general():
    """3Q-VAS method test"""
    mth = ThreeQ_VAS(channels=["87Rb"], spectral_dimensions=[{}, {}])
    assert mth.name == "ThreeQ_VAS"
    assert mth.description == "Simulate a 3Q variable-angle spinning spectrum."
    assert mth.spectral_dimensions[0].events[0].transition_query == TransitionQuery(
        P={"channel-1": [[-3]]}, D={"channel-1": [[0]]}
    )
    assert mth.spectral_dimensions[1].events[0].transition_query == TransitionQuery(
        P={"channel-1": [[-1]]}, D={"channel-1": [[0]]}
    )
    assert ThreeQ_VAS.parse_dict_with_units(mth.json()) == mth

    assert np.allclose(mth.affine_matrix, [0.5625, 0.4375, 0.0, 1.0])

    serialize = mth.json()
    _ = serialize.pop("affine_matrix")
    assert serialize == {
        "channels": ["87Rb"],
        "description": "Simulate a 3Q variable-angle spinning spectrum.",
        "magnetic_flux_density": "9.4 T",
        "name": "ThreeQ_VAS",
        "rotor_angle": "0.955316618 rad",
        "rotor_frequency": "1000000000000.0 Hz",
        "spectral_dimensions": [
            {
                "count": 1024,
                "reference_offset": "0.0 Hz",
                "spectral_width": "25000.0 Hz",
            },
            {
                "count": 1024,
                "reference_offset": "0.0 Hz",
                "spectral_width": "25000.0 Hz",
            },
        ],
    }


def test_5Q_VAS_general():
    """5Q-VAS method test"""
    mth = FiveQ_VAS(channels=["17O"], spectral_dimensions=[{}, {}])

    assert mth.name == "FiveQ_VAS"
    assert mth.description == "Simulate a 5Q variable-angle spinning spectrum."
    assert mth.spectral_dimensions[0].events[0].transition_query == TransitionQuery(
        P={"channel-1": [[-5]]}, D={"channel-1": [[0]]}
    )
    assert mth.spectral_dimensions[1].events[0].transition_query == TransitionQuery(
        P={"channel-1": [[-1]]}, D={"channel-1": [[0]]}
    )
    assert FiveQ_VAS.parse_dict_with_units(mth.json()) == mth

    assert np.allclose(
        mth.affine_matrix, [0.3243243243243243, 0.6756756756756757, 0.0, 1.0]
    )

    serialize = mth.json()
    _ = serialize.pop("affine_matrix")

    assert serialize == {
        "channels": ["17O"],
        "description": "Simulate a 5Q variable-angle spinning spectrum.",
        "magnetic_flux_density": "9.4 T",
        "name": "FiveQ_VAS",
        "rotor_angle": "0.955316618 rad",
        "rotor_frequency": "1000000000000.0 Hz",
        "spectral_dimensions": [
            {
                "count": 1024,
                "reference_offset": "0.0 Hz",
                "spectral_width": "25000.0 Hz",
            },
            {
                "count": 1024,
                "reference_offset": "0.0 Hz",
                "spectral_width": "25000.0 Hz",
            },
        ],
    }


def test_7Q_VAS_general():
    """7Q-VAS method test"""
    mth = SevenQ_VAS(channels=["51V"], spectral_dimensions=[{}, {}])

    assert mth.name == "SevenQ_VAS"
    assert mth.description == "Simulate a 7Q variable-angle spinning spectrum."
    assert mth.spectral_dimensions[0].events[0].transition_query == TransitionQuery(
        P={"channel-1": [[-7]]}, D={"channel-1": [[0]]}
    )
    assert mth.spectral_dimensions[1].events[0].transition_query == TransitionQuery(
        P={"channel-1": [[-1]]}, D={"channel-1": [[0]]}
    )
    assert SevenQ_VAS.parse_dict_with_units(mth.json()) == mth

    assert np.allclose(mth.affine_matrix, [0.2184466, 0.7815534, 0.0, 1.0])

    serialize = mth.json()
    _ = serialize.pop("affine_matrix")

    assert serialize == {
        "channels": ["51V"],
        "description": "Simulate a 7Q variable-angle spinning spectrum.",
        "magnetic_flux_density": "9.4 T",
        "name": "SevenQ_VAS",
        "rotor_angle": "0.955316618 rad",
        "rotor_frequency": "1000000000000.0 Hz",
        "spectral_dimensions": [
            {
                "count": 1024,
                "reference_offset": "0.0 Hz",
                "spectral_width": "25000.0 Hz",
            },
            {
                "count": 1024,
                "reference_offset": "0.0 Hz",
                "spectral_width": "25000.0 Hz",
            },
        ],
    }



# -*- coding: utf-8 -*-
import os

from mrsimulator import load
from mrsimulator import save
from mrsimulator import signal_processing as sp
from mrsimulator import Simulator
from mrsimulator import Site
from mrsimulator import SpinSystem
from mrsimulator.methods import BlochDecaySpectrum

# mrsimulator save and load test


def setup():
    site = Site(isotope="1H", shielding_symmetric={"zeta": -100, "eta": 0.3})
    spin_sys = SpinSystem(sites=[site, site], abundance=45)

    method = BlochDecaySpectrum(channels=["1H"])
    sim = Simulator(spin_systems=[spin_sys, spin_sys], methods=[method, method])

    sim.run(method_index=0)

    processor = sp.SignalProcessor(
        operations=[
            sp.IFFT(),
            sp.apodization.Exponential(FWHM="2500 Hz"),
            sp.FFT(),
            sp.Scale(factor=20),
        ]
    )
    processors = [processor] * 2

    return sim, processors


def test_save():
    sim, processors = setup()

    assert sim.methods[0].simulation is not None
    assert sim.methods[1].simulation is None

    save("test.mrsim", simulator=sim, signal_processors=processors, params=None)


def test_load():
    sim_r, processors_r, report_r = load("test.mrsim")

    sim, processors = setup()
    sim_r.methods[0].simulation = None
    sim.methods[0].simulation = None
    assert sim_r == sim
    assert processors_r == processors
    assert report_r is None

    os.remove("test.mrsim")



# -*- coding: utf-8 -*-
import csdmpy as cp
import numpy as np
from mrsimulator import signal_processing as sp

from .test_signal_processing import generate_data
from .test_signal_processing import setup_read_write


def test_baseline_constant_offset():
    data_in = generate_data()
    PS_0 = [sp.baseline.ConstantOffset(offset=10)]
    operator = sp.SignalProcessor(operations=PS_0)
    data_out = operator.apply_operations(data=data_in.copy())
    _, y0, y1, y2 = data_in.to_list()
    _, y0_, y1_, y2_ = data_out.to_list()

    for in_, out_ in zip([y0, y1, y2], [y0_, y1_, y2_]):
        assert np.allclose(out_.max() - in_.max(), 10), "Offset failed"

    py_dict = {
        "function": "baseline",
        "type": "ConstantOffset",
        "offset": 10.0,
    }
    setup_read_write(PS_0[0], py_dict, sp.baseline.ConstantOffset)


def test_baseline_polynomial():
    data_in = generate_data()
    data_in.dimensions[0] *= cp.ScalarQuantity("1 ms")
    x_c = data_in.dimensions[0].coordinates.to("s").value

    # zeroth order
    PS_0 = [sp.baseline.Polynomial(x0="10")]
    data_out = sp.SignalProcessor(operations=PS_0).apply_operations(data=data_in.copy())

    _, y0, y1, y2 = data_in.to_list()
    _, y0_, y1_, y2_ = data_out.to_list()

    for in_, out_ in zip([y0, y1, y2], [y0_, y1_, y2_]):
        assert np.allclose(out_.max() - in_.max(), 10), "Offset failed"

    py_dict = {
        "function": "baseline",
        "type": "Polynomial",
        "x0": "10.0",
        "x1": "0",
        "x2": "0",
        "x3": "0",
        "dim_index": 0,
    }
    setup_read_write(PS_0[0], py_dict, sp.baseline.Polynomial)

    # first order
    PS_0 = [sp.baseline.Polynomial(x0="30", x1="1 Hz")]
    data_out = sp.SignalProcessor(operations=PS_0).apply_operations(data=data_in.copy())

    _, y0, y1, y2 = data_in.to_list()
    _, y0_, y1_, y2_ = data_out.to_list()

    for in_, out_ in zip([y0, y1, y2], [y0_, y1_, y2_]):
        assert np.allclose(out_, in_ + 30 + x_c), "Polynomial 1st order failed"

    py_dict = {
        "function": "baseline",
        "type": "Polynomial",
        "x0": 30.0,
        "x1": "1.0 Hz",
        "x2": 0,
        "x3": 0,
        "dim_index": 0,
    }
    setup_read_write(PS_0[0], py_dict, sp.baseline.Polynomial)

    # second order
    PS_0 = [sp.baseline.Polynomial(x0="1", x2="1 Hz^2")]
    data_out = sp.SignalProcessor(operations=PS_0).apply_operations(data=data_in.copy())

    _, y0, y1, y2 = data_in.to_list()
    _, y0_, y1_, y2_ = data_out.to_list()

    for in_, out_ in zip([y0, y1, y2], [y0_, y1_, y2_]):
        assert np.allclose(out_, in_ + 1 + x_c ** 2), "Polynomial 2nd order failed"

    py_dict = {
        "function": "baseline",
        "type": "Polynomial",
        "x0": 1.0,
        "x1": 0,
        "x2": "1.0 Hz2",
        "x3": 0,
        "dim_index": 0,
    }
    setup_read_write(PS_0[0], py_dict, sp.baseline.Polynomial)

    # third order
    PS_0 = [sp.baseline.Polynomial(x0="10", x3="2 Hz^3", x1="13.1 Hz")]
    data_out = sp.SignalProcessor(operations=PS_0).apply_operations(data=data_in.copy())

    _, y0, y1, y2 = data_in.to_list()
    _, y0_, y1_, y2_ = data_out.to_list()

    for in_, out_ in zip([y0, y1, y2], [y0_, y1_, y2_]):
        assert np.allclose(
            out_, in_ + 10 + 13.1 * x_c + 2 * x_c ** 3
        ), "Polynomial 3rd order failed"

    # py_dict = {
    #     "function": "baseline",
    #     "type": "Polynomial",
    #     "x0": 10.0,
    #     "x1": "1.0 Hz",
    #     "x2": 0,
    #     "x3": "2.0 Hz3",
    #     "dim_index": 0,
    # }
    # setup_read_write(PS_0[0], py_dict, sp.baseline.Polynomial)



# -*- coding: utf-8 -*-
"""Base SpinSystem class."""
from copy import deepcopy
from typing import ClassVar
from typing import Dict
from typing import List
from typing import Union

import numpy as np
from mrsimulator.base_model import get_zeeman_states
from mrsimulator.transition import Transition
from mrsimulator.transition.pathway import TransitionList
from mrsimulator.transition.pathway import TransitionPathway
from mrsimulator.utils.parseable import Parseable
from pydantic import Field
from pydantic import validator

from .coupling import Coupling
from .isotope import ISOTOPE_DATA
from .site import Site
from .zeemanstate import ZeemanState

__author__ = "Deepansh Srivastava"
__email__ = "srivastava.89@osu.edu"


class SpinSystem(Parseable):
    """Base class representing an isolated spin system containing multiple sites and
    couplings amongst them.

    .. rubric:: Attribute Documentation

    Attributes
    ----------

    sites: list of :ref:`site_api` or equivalent dict objects (optional)
        A list of :ref:`site_api` or equivalent dict objects within the spin system.
        Each site object represents single-site nuclear spin interaction (nuclear
        shielding and EFG) tensor parameters. The default value is an empty list.

        Example
        -------

        >>> sys1 = SpinSystem()
        >>> sys1.sites = [Site(isotope="17O"), Site(isotope="1H")]
        >>> # or equivalently
        >>> sys1.sites = [{"isotope": "17O"}, {"isotope": "1H"}]

    couplings: list of :ref:`coupling_api` or equivalent dict objects (optional)
        A list of :ref:`coupling_api` or equivalent dict objects within the spin
        system. Each coupling object represents two-site spin interaction (J-coupling
        and Dipolar) tensor parameters. The default value is an empty list.

        Example
        -------

        >>> sys1 = SpinSystem()
        >>> sys1.couplings = [
        ...     Coupling(site_index=[0, 1], isotropic_j=10.1),
        ...     Coupling(site_index=[2, 1], dipolar={"D": 1500})
        ... ]
        >>> # or equivalently
        >>> sys1.couplings = [
        ...     {"site_index": [0, 1], "isotropic_j": 10.1},
        ...     {"site_index": [2, 1], "dipolar": {"D": 1500}}
        ... ]

    abundance: float (optional).
        The abundance of the spin system in units of %. The default value is 100. The
        value of this attribute is useful when multiple spin systems are present.

        Example
        -------

        >>> sys1.abundance = 10

    name: str (optional).
        The value is the name or id of the spin system. The default value is None.

        Example
        -------

        >>> sys1.name = "1H-17O-0"
        >>> print(sys1.name)
        1H-17O-0

    label: str (optional).
        The value is a label for the spin system. The default value is None.

        Example
        -------

        >>> sys1.label = "Heteronuclear spin system"
        >>> print(sys1.label)
        Heteronuclear spin system

    description: str (optional).
        The value is a description of the spin system. The default value is None.

        Example
        -------

        >>> sys1.description = "A test for the spin system"
        >>> print(sys1.description)
        A test for the spin system

    transition_pathways: list of :ref:`transition_pathway_api` (optional).
        A list of :ref:`transition_pathway_api` or equivalent dict objects. Each
        transition pathway is a list of :ref:`transition_api` objects. The resulting
        spectrum is a sum of the resonances arising from individual transition pathways.
        The default value is None.

        Example
        -------
        >>> sys1.transition_pathways = [
        ...     [
        ...         {"initial": [-2.5, 0.5], "final": [2.5, 0.5]},
        ...         {"initial": [0.5, 0.5], "final": [-0.5, 0.5]}
        ...     ]
        ... ]
        >>> print(sys1.transition_pathways)
        [|2.5, 0.5⟩⟨-2.5, 0.5| ⟶ |-0.5, 0.5⟩⟨0.5, 0.5|]

        .. note::
            From any given spin system, the list of relevant transition pathways is
            determined by the NMR method. For example, consider a single site
            I=3/2 spin system. For this system, a Bloch decay spectrum method will
            select three transition pathways, one corresponding to the central and two
            to the satellite transitions. On the other hand, a Bloch decay central
            transition selective method will only select one transition pathway,
            corresponding to the central transition.

            Since the spin system is independent of the NMR method, the value of this
            attribute is, therefore, transient. You may use this attribute to override
            the default transition pathway query selection criterion of the NMR method
            objects.

            **Only use this attribute if you know what you are doing.**

            At times, this attribute may provide a significant improvement in the
            performance, especially in iterative algorithms, such as the least-squares
            algorithm, where a one-time transition pathway query is sufficient.
            Repeated queries for the transition pathways will add significant overhead
            to the computation.

            .. seealso::
                `Fitting example
                <./../examples/Fitting/plot_2_mrsimFitExample_O17.html>`_
    """

    name: str = None
    label: str = None
    description: str = None
    sites: Union[List[Site], np.ndarray] = []
    couplings: Union[List[Coupling], np.ndarray] = None
    abundance: float = Field(default=100.0, ge=0.0, le=100.0)
    transition_pathways: List = None

    property_unit_types: ClassVar = {"abundance": "dimensionless"}
    property_default_units: ClassVar = {"abundance": "pct"}
    property_units: Dict = {"abundance": "pct"}

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    @validator("transition_pathways")
    def transition_pathways_must_include_transition(cls, v, values):
        if v is None:
            return v
        return [TransitionPathway(item) for item in v]

    @validator("sites")
    def check_sites(cls, v, values):
        if isinstance(v, np.ndarray):
            if not np.all([isinstance(item, Site) for item in v]):
                raise ValueError("All entries must be of type `Site`.")
        return list(v)

    @validator("couplings")
    def check_couplings(cls, v, values):
        if isinstance(v, np.ndarray):
            if not np.all([isinstance(item, Coupling) for item in v]):
                raise ValueError("All entries must be of type `Coupling`.")
        return list(v)

    def get_isotopes(self, spin_I: float = None, symbol: bool = False) -> list:
        """An ordered list of :ref:`isotope_api` objects from the sites within the spin
        system corresponding to the given value of spin quantum number `I`. If `I` is
        None, a list of all Isotope objects is returned instead.

        Args:
            float spin_I: An optional spin quantum number. The valid inputs are the
                multiples of 0.5.
            bool symbol: If true, return a list of str with isotope symbols.

        Returns:
            A list of :ref:`isotope_api` objects.

        Example
        -------

        >>> spin_systems.get_isotopes() # three spin systems
        [Isotope(symbol='13C'), Isotope(symbol='1H'), Isotope(symbol='27Al')]
        >>> spin_systems.get_isotopes(symbol=True) # three spin systems
        ['13C', '1H', '27Al']

        >>> spin_systems.get_isotopes(spin_I=0.5) # isotopes with I=0.5
        [Isotope(symbol='13C'), Isotope(symbol='1H')]
        >>> spin_systems.get_isotopes(spin_I=0.5, symbol=True) # isotopes with I=0.5
        ['13C', '1H']

        >>> spin_systems.get_isotopes(spin_I=1.5) # isotopes with I=1.5
        []

        >>> spin_systems.get_isotopes(spin_I=2.5) # isotopes with I=2.5
        [Isotope(symbol='27Al')]
        >>> spin_systems.get_isotopes(spin_I=2.5, symbol=True) # isotopes with I=2.5
        ['27Al']
        """
        isotope_list = allowed_isotopes(spin_I)
        if not symbol:
            return [
                site.isotope
                for site in self.sites
                if site.isotope.symbol in isotope_list
            ]

        return [
            site.isotope.symbol
            for site in self.sites
            if site.isotope.symbol in isotope_list
        ]

    def _zeeman_energy_states(self) -> np.ndarray:
        """Return the energy states as a Numpy array where the axis 0 is the number of
        energy states, and axis 1 is the spin quantum numbers. The spin quantum numbers
        are ordered based on the order of the sites within the spin system.
        """
        return get_zeeman_states(self)

    def zeeman_energy_states(self) -> list:
        r"""Return a list of all :ref:`zeeman_api` objects of the spin system, where the
        energy states are represented by a list of quantum numbers,

        .. math::
            |\Psi⟩ = [m_1, m_2,.. m_n],

        where :math:`m_i` is the quantum number associated with the :math:`i^\text{th}`
        site within the spin system, and :math:`\Psi` is the energy state.

        Example
        -------

        >>> spin_system_1H_13C.zeeman_energy_states()  # four energy level system.
        [|-0.5, -0.5⟩, |-0.5, 0.5⟩, |0.5, -0.5⟩, |0.5, 0.5⟩]

        Returns
            A list of :ref:`zeeman_api` objects.
        """
        states = self._zeeman_energy_states()
        return [ZeemanState(len(self.sites), *item) for item in states]

    def _all_transitions(self) -> np.ndarray:
        """All transitions from a spin system as a Numpy array of shape (M, 2, N),
        where M is the number of transitions, and N is the number of sites in the
        spin system. The second axis is of length 2, where the entries at T[:, 0, :]
        are the initial energy states, and the entries at T[:, 1, :], the corresponding
        final energy states of the spin transitions.
        """
        energy_states = self._zeeman_energy_states()
        s = energy_states.shape[0]
        lst = np.arange(s)
        indexes = np.asarray(np.meshgrid(lst, lst)).T.reshape(-1, 2)
        return energy_states[indexes]

    def all_transitions(self) -> TransitionList:
        """Returns a list of all possible spin :ref:`transition_api` objects in the
        given spin system.

        Example
        -------

        >>> spin_system_1H_13C.all_transitions()  # 16 two energy level transitions
        [|-0.5, -0.5⟩⟨-0.5, -0.5|,
        |-0.5, 0.5⟩⟨-0.5, -0.5|,
        |0.5, -0.5⟩⟨-0.5, -0.5|,
        |0.5, 0.5⟩⟨-0.5, -0.5|,
        |-0.5, -0.5⟩⟨-0.5, 0.5|,
        |-0.5, 0.5⟩⟨-0.5, 0.5|,
        |0.5, -0.5⟩⟨-0.5, 0.5|,
        |0.5, 0.5⟩⟨-0.5, 0.5|,
        |-0.5, -0.5⟩⟨0.5, -0.5|,
        |-0.5, 0.5⟩⟨0.5, -0.5|,
        |0.5, -0.5⟩⟨0.5, -0.5|,
        |0.5, 0.5⟩⟨0.5, -0.5|,
        |-0.5, -0.5⟩⟨0.5, 0.5|,
        |-0.5, 0.5⟩⟨0.5, 0.5|,
        |0.5, -0.5⟩⟨0.5, 0.5|,
        |0.5, 0.5⟩⟨0.5, 0.5|]

        Returns
            A list of :ref:`transition_api` objects.
        """
        transitions = self._all_transitions()
        return TransitionList(
            [
                Transition(initial=item[0].tolist(), final=item[1].tolist())
                for item in transitions
            ]
        )

    @classmethod
    def parse_dict_with_units(cls, py_dict: dict):
        """Parse physical quantities from a dictionary representation of the SpinSystem
        object, where the physical quantity is expressed as a string with a number and
        a unit.

        Args:
            dict py_dict: A required python dict object.

        Returns:
            :ref:`spin_sys_api` object.

        Example
        -------

        >>> spin_system_dict = {
        ...     "sites": [{
        ...         "isotope":"13C",
        ...         "isotropic_chemical_shift": "20 ppm",
        ...         "shielding_symmetric": {
        ...             "zeta": "10 ppm",
        ...             "eta": 0.5
        ...         }
        ...     }]
        ... }
        >>> spin_system_1 = SpinSystem.parse_dict_with_units(spin_system_dict)
        """
        py_dict_copy = deepcopy(py_dict)
        if "sites" in py_dict_copy:
            py_dict_copy["sites"] = [
                Site.parse_dict_with_units(s) for s in py_dict_copy["sites"]
            ]

        if "couplings" in py_dict_copy:
            py_dict_copy["couplings"] = [
                Coupling.parse_dict_with_units(s) for s in py_dict_copy["couplings"]
            ]

        return super().parse_dict_with_units(py_dict_copy)

    # Deprecated
    # def to_freq_dict(self, B0: float) -> dict:
    #     """
    #     Serialize the SpinSystem object to a JSON compliant python dictionary object,
    #     where the attribute value is a numbers expressed in the attribute's default
    #     unit. The default unit for the attributes with respective dimensionalities
    #     are:

    #     - frequency: `Hz`
    #     - angle: `rad`

    #     Args:
    #         float B0: A required macroscopic magnetic flux density in units of T.

    #     Return:
    #         A python dict

    #     Example
    #     -------

    #     >>> pprint(spin_system_1.to_freq_dict(B0=9.4))
    #     {"abundance": 100,
    #      "description": None,
    #      "label": None,
    #      "name": None,
    #      "sites": [{"description": None,
    #                 "isotope": "13C",
    #                 "isotropic_chemical_shift": -2013.1791999999998,
    #                 "label": None,
    #                 "name": None,
    #                 "quadrupolar": None,
    #                 "shielding_antisymmetric": None,
    #                 "shielding_symmetric": {"alpha": None,
    #                                         "beta": None,
    #                                         "eta": 0.5,
    #                                         "gamma": None,
    #                                         "zeta": -1006.5895999999999}}],
    #      "transition_pathways": None}
    #     """
    #     temp_dict = self.dict()
    #     temp_dict["sites"] = [site.to_freq_dict(B0) for site in self.sites]
    #     temp_dict.pop("property_units")
    #     return temp_dict


def allowed_isotopes(spin_I: float = None) -> list:
    """List of NMR active isotopes currently supported in ``mrsimulator``.

    Args:
        float spin_I: Optional spin quantum number. The valid values are multiples
            of 0.5. The default is None.

    Returns:
        A list of all isotopes supported in ``mrsimulator`` with the given spin
        quantum number `I`. If the spin is unspecified or None, a list of all
        allowed isotopes is returned instead.
    """
    if spin_I is None:
        return list(ISOTOPE_DATA.keys())
    return list(
        {
            isotope
            for isotope, data in ISOTOPE_DATA.items()
            if data["spin"] == int(2 * spin_I)
        }
    )



# -*- coding: utf-8 -*-
import platform
import sys
from os.path import abspath
from os.path import dirname
from os.path import exists
from os.path import join
from os import environ

from setuptools import Extension
from setuptools import find_packages
from setuptools import setup

import numpy as np
import numpy.distutils.system_info as sysinfo

from settings import use_accelerate
from settings import use_openblas

try:
    from Cython.Build import cythonize

    USE_CYTHON = True
except ImportError:
    USE_CYTHON = False


def message(lib, env, command, key):
    arg = f"{key} {lib}" if key != "" else f"{lib}"
    print(f"Error: Libraries not found - {lib}")
    print(f"Please install '{lib}' from {env} with:\n\t{command} install {arg}")
    sys.exit(1)


class Setup:
    __slots__ = [
        "include_dirs",
        "library_dirs",
        "libraries",
        "extra_compile_args",
        "extra_link_args",
    ]

    def __init__(self):
        self.libraries = []
        self.include_dirs = []
        self.library_dirs = []
        self.extra_compile_args = []
        self.extra_link_args = []

    def check_valid_path(self, pathlist):
        return [pth for pth in pathlist if exists(pth)]

    def check_if_lib_exists(self, lib):
        return np.any([exists(join(pth, lib)) for pth in self.library_dirs])

    def check_if_header_exists(self, header):
        return np.any([exists(join(pth, header)) for pth in self.include_dirs])

    def conda_setup_for_windows(self):
        self.libraries += ["fftw3", "openblas"]
        self.extra_compile_args = ["/DUSE_OPENBLAS"]

        print(sys.version)
        loc = dirname(sys.executable)
        if "conda" not in loc:
            return

        print("Found Python installation:", loc)
        self.include_dirs += self.check_valid_path(
            [
                join(loc, "Library", "include", "fftw"),
                join(loc, "Library", "include", "openblas"),
                join(loc, "Library", "include"),
                join(loc, "include"),
            ]
        )
        self.library_dirs += self.check_valid_path([join(loc, "Library", "lib")])
        environ["MRSIM_LIB"] = str(join(loc, "Library", "lib"))
        self.on_exit_message("openblas.lib", "fftw3.lib")

    def conda_setup_for_unix(self):
        loc = dirname(sys.executable)
        print("Found Python installation:", loc)

        self.include_dirs += self.check_valid_path([join(loc, "include")])
        self.library_dirs += self.check_valid_path([join(loc, "lib")])
        self.extra_compile_args = ["-O3", "-ffast-math", "-DUSE_OPENBLAS"]
        self.libraries += ["fftw3", "openblas"]

    def on_exit_message(self, blas_lib, fftw_lib):
        found_blas = self.check_if_lib_exists(blas_lib)
        found_fftw = self.check_if_lib_exists(fftw_lib)

        cmd_list = ["conda", "conda", "-c conda-forge"]
        if not found_blas and not found_fftw:
            message("openblas fftw", *cmd_list)

        if not found_blas:
            message("openblas", *cmd_list)

        if not found_fftw:
            message("fftw", *cmd_list)

    # def numpy_default_blas(self):
    #     opt_info = np.__config__.blas_opt_info
    #     if opt_info == {}:
    #         return

    #     if "pthread" in opt_info["libraries"]:
    #         opt_info["libraries"].remove("pthread")

    #     libs = opt_info["libraries"]
    #     print(f"Linking mrsimulator with the default numpy blas: {libs}")

    #     self.include_dirs += opt_info["include_dirs"]
    #     self.library_dirs += opt_info["library_dirs"]
    #     self.libraries += libs
    #     self.BLAS_FOUND = True

    def mkl_blas_info(self):
        mkl_info = np.__config__.blas_mkl_info
        if mkl_info == {}:
            print("Please enable mkl for numpy before proceeding.")
            message("mkl mkl-include", "pip", "pip", "")

        self.include_dirs += mkl_info["include_dirs"]
        self.library_dirs += mkl_info["library_dirs"]
        self.libraries += mkl_info["libraries"]

        if not self.check_if_header_exists("mkl.h"):
            print("mkl header file not found.")
            message("mkl-include", "pip", "pip", "")

        print("Linking mrsimulator with the mkl blas.")
        self.extra_compile_args += ["-DUSE_MKL", "/DUSE_MKL"]


class WindowsSetup(Setup):
    def __init__(self):
        super().__init__()

        self.extra_link_args += ["-Wl"]
        self.extra_compile_args = ["-DFFTW_DLL"]

        # if use_mkl:
        #     self.mkl_blas_info()

        self.conda_setup_for_windows()


class LinuxSetup(Setup):
    def __init__(self):
        super().__init__()
        self.extra_compile_args = [
            "-O3",
            "-ffast-math",
            # "-msse4.2",
            # "-ftree-vectorize",
            # "-fopt-info-vec-all",
            # "-fopt-info-vec-optimized",
            # "-mavx",
            "-g",
            "-DUSE_OPENBLAS",
        ]
        self.extra_link_args += ["-lm"]
        self.include_dirs += [
            "/usr/include/",
            "/usr/include/openblas",
            "/usr/include/x86_64-linux-gnu/",
        ]

        self.library_dirs += ["/usr/lib64/", "/usr/lib/", "/usr/lib/x86_64-linux-gnu/"]
        self.libraries += ["openblas", "fftw3"]
        openblas_info = sysinfo.get_info("openblas")
        fftw3_info = sysinfo.get_info("fftw3")

        if openblas_info == {} and fftw3_info == {}:
            self.message("openblas-devel fftw-devel", "libopenblas-dev libfftw3-dev")

        if openblas_info == {}:
            self.message("openblas-devel", "libopenblas-dev")

        if fftw3_info == {}:
            self.message("fftw-devel", "libfftw3-dev")

        self.get_location(openblas_info)
        self.get_location(fftw3_info)

    def get_location(self, dict_info):
        for item in self.__slots__:
            if item in dict_info.keys():
                getattr(self, item).extend(dict_info[item])

    def message(self, lib_centos, lib_ubuntu):
        stat = f"yum install {lib_centos}"
        print(f"For CentOS users\n\t{stat}")
        stat = f"sudo apt-get update\n\tsudo apt-get install {lib_ubuntu}"
        print(f"For Ubuntu users\n\t{stat}")
        sys.exit(1)


class MacOSSetup(Setup):
    def __init__(self):
        super().__init__()
        self.extra_compile_args = [
            "-O3",
            "-ffast-math",
            # "-Rpass=loop-vectorize",
            # "-Rpass-missed=loop-vectorize",
            # "-Rpass-analysis=loop-vectorize",
            "-fvectorize",
            "-fcommon",
        ]
        self.extra_link_args += ["-lm"]

        # Blas
        if use_accelerate:
            self.accelerate_info()

        if use_openblas:
            self.openblas_info()

        # if use_mkl:
        #     self.mkl_blas_info()

        # FFTW
        self.fftw_info()

    def accelerate_info(self):
        """Apple's Accelerate framework for BLAS"""
        acc_info = sysinfo.get_info("accelerate")
        for item in ["extra_compile_args", "extra_link_args"]:
            if item in acc_info:
                self.extra_compile_args += acc_info[item]
        print("Linking mrsimulator with Apple accelerate library.")
        self.extra_compile_args += ["-DUSE_ACCELERATE"]

    def openblas_info(self):
        """openblas includes and lib are for brew installation"""
        blas_include_dir = "/usr/local/opt/openblas/include"
        blas_library_dir = "/usr/local/opt/openblas/lib"
        blas_library = "openblas"

        if not exists(blas_include_dir):
            message("openblas", "homebrew", "brew", "")

        print("Linking mrsimulator with openblas library.")
        self.include_dirs += [blas_include_dir]
        self.library_dirs += [blas_library_dir]
        self.libraries += [blas_library]
        self.extra_compile_args += ["-DUSE_OPENBLAS"]

    # def mkl_blas_info(self):
    #     mkl_info = np.__config__.blas_mkl_info
    #     if mkl_info == {}:
    #         print("Please enable mkl for numpy before proceeding.")
    #         message("mkl mkl-include", "pip", "pip", "")

    #     if not self.check_if_file_exists("mkl.h"):
    #         print("mkl header file not found.")
    #         message("mkl-include", "pip", "pip", "")

    #     self.include_dirs += mkl_info["include_dirs"]
    #     self.library_dirs += mkl_info["library_dirs"]
    #     self.libraries += mkl_info["libraries"]
    #     self.extra_compile_args += ["-DUSE_MKL"]

    def fftw_info(self):
        """fftw includes and lib are for brew installation"""
        fftw_include_dir = "/usr/local/opt/fftw/include"
        fftw_library_dir = "/usr/local/opt/fftw/lib"
        fftw_library = "fftw3"

        if not exists(fftw_include_dir):
            message("fftw", "homebrew", "brew", "")

        print("Linking mrsimulator with fftw library.")
        self.include_dirs += [fftw_include_dir]
        self.library_dirs += [fftw_library_dir]
        self.libraries += [fftw_library]


# get the version from file
python_version = sys.version_info
py_version = ".".join([str(i) for i in python_version[:3]])
print("Using python version", py_version)
if python_version.major != 3 and python_version.minor < 6:
    print(f"Python>=3.6 is required for the setup. You are using version {py_version}")
    sys.exit(1)

with open("src/mrsimulator/__init__.py", "r") as f:
    for line in f.readlines():
        if "__version__" in line:
            before_keyword, keyword, after_keyword = line.partition("=")
            version = after_keyword.strip()[1:-1]
            print("mrsimulator version ", version)
            break

module_dir = dirname(abspath(__file__))
data_files = []
numpy_include = np.get_include()

if sys.platform.startswith("win"):
    win = WindowsSetup()

if platform.system() == "Darwin":
    win = MacOSSetup()

if platform.system() == "Linux":
    win = LinuxSetup()

extra_link_args = list(set(win.extra_link_args))
extra_compile_args = list(set(win.extra_compile_args))
library_dirs = list(set(win.library_dirs))
include_dirs = list(set(win.include_dirs))
libraries = list(set(win.libraries))

# other include paths
include_dirs += ["src/c_lib/include/", numpy_include]

# print info
print(include_dirs)
print(library_dirs)
print(libraries)
print(extra_compile_args)
print(extra_link_args)

source = [
    "src/c_lib/lib/angular_momentum.c",
    "src/c_lib/lib/interpolation.c",
    "src/c_lib/lib/method.c",
    "src/c_lib/lib/mrsimulator.c",
    "src/c_lib/lib/octahedron.c",
    "src/c_lib/lib/frequency_averaging.c",
    "src/c_lib/lib/schemes.c",
    "src/c_lib/lib/simulation.c",
]

ext = ".pyx" if USE_CYTHON else ".c"

# method
ext_modules = [
    Extension(
        name="mrsimulator.base_model",
        sources=[*source, "src/c_lib/base/base_model" + ext],
        include_dirs=include_dirs,
        language="c",
        libraries=libraries,
        library_dirs=library_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]

# tests
ext_modules += [
    Extension(
        name="mrsimulator.tests.tests",
        sources=[*source, "src/c_lib/test/test" + ext],
        include_dirs=include_dirs,
        language="c",
        libraries=libraries,
        library_dirs=library_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]

# sandbox
ext_modules += [
    Extension(
        name="mrsimulator.sandbox",
        sources=[*source, "src/c_lib/sandbox/sandbox" + ext],
        include_dirs=include_dirs,
        language="c",
        libraries=libraries,
        library_dirs=library_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]

if USE_CYTHON:
    ext_modules = cythonize(ext_modules, language_level=3)

extras = {"all": ["matplotlib>=3.3.3"]}

description = "A python toolbox for simulating fast real-time solid-state NMR spectra."
setup(
    name="mrsimulator",
    version=version,
    description=description,
    long_description=open(join(module_dir, "README.md")).read(),
    long_description_content_type="text/markdown",
    author="Deepansh J. Srivastava",
    author_email="srivastava.89@osu.edu",
    python_requires=">=3.6",
    url="https://github.com/DeepanshS/MRsimulator/",
    packages=find_packages("src"),
    package_dir={"": "src"},
    setup_requires=["numpy>=1.17"],
    install_requires=[
        "numpy>=1.17",
        "csdmpy>=0.4",
        "pydantic>=1.0",
        "monty>=2.0.4",
        "typing-extensions>=3.7",
        "psutil>=5.4.8",
        "joblib>=1.0.0",
        "pandas>=1.1.3",
        "lmfit>=1.0.2",
    ],
    extras_require=extras,
    ext_modules=ext_modules,
    include_package_data=True,
    zip_safe=False,
    license="BSD-3-Clause",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: C",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Education",
        "Topic :: Scientific/Engineering :: Chemistry",
    ],
)



# -*- coding: utf-8 -*-
import numpy as np
from mrsimulator.spin_system.tensors import SymmetricTensor

from .utils import get_Haeberlen_components
from .utils import get_principal_components
from .utils import x_y_from_zeta_eta

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


def _czjzek_random_distribution_tensors(sigma, n):
    r"""Czjzek random distribution model.

    Args:
        float sigma: The standard deviation of the five-dimensional multi-variate normal
            distribution.
        int n: Number of samples drawn from the Czjzek random distribution model.

    Description
    -----------

    U is an array of the coordinates randomly drawn from an uncorrelated
    five-dimensional multivariate normal distribution with standard deviation `sigma`
    and zero mean.

    The components of the traceless second-rank symmetric cartesian tensor, S_ij,
    follows

    Sxx = sqrt(3) * U5 - U1

    Sxy = Syx = sqrt(3) * U4

    Syy = -sqrt(3) * U5 - U1

    Sxz = Szx = sqrt(3) * U2

    Szz = 2 * U1

    Syz = Szy = sqrt(3) * U3
    """

    # The random sampling U1, U2, ... U5
    U1 = np.random.normal(0.0, sigma, n)

    sqrt_3_sigma = np.sqrt(3) * sigma
    sqrt_3_U2 = np.random.normal(0.0, sqrt_3_sigma, n)
    sqrt_3_U3 = np.random.normal(0.0, sqrt_3_sigma, n)
    sqrt_3_U4 = np.random.normal(0.0, sqrt_3_sigma, n)
    sqrt_3_U5 = np.random.normal(0.0, sqrt_3_sigma, n)

    # Create N random tensors
    tensors = np.zeros((n, 3, 3))  # n x 3 x 3 tensors

    tensors[:, 0, 0] = sqrt_3_U5 - U1  # xx
    tensors[:, 0, 1] = sqrt_3_U4  # xy
    tensors[:, 0, 2] = sqrt_3_U2  # xz

    tensors[:, 1, 0] = sqrt_3_U4  # yx
    tensors[:, 1, 1] = -sqrt_3_U5 - U1  # yy
    tensors[:, 1, 2] = sqrt_3_U3  # yz

    tensors[:, 2, 0] = sqrt_3_U2  # zx
    tensors[:, 2, 1] = sqrt_3_U3  # zy
    tensors[:, 2, 2] = 2 * U1  # zz

    return tensors


class AbstractDistribution:
    def pdf(self, pos, size: int = 400000):
        """Generates a probability distribution function by binning the random
        variates of length size onto the given grid system.

        Args:
            pos: A list of coordinates along the two dimensions given as NumPy arrays.
            size: The number of random variates drawn in generating the pdf. The default
                is 400000.

        Returns:
            A list of x and y coordinates and the corresponding amplitudes.

        Example:
            >>> import numpy as np
            >>> cq = np.arange(50) - 25
            >>> eta = np.arange(21)/20
            >>> Cq_dist, eta_dist, amp = cz_model.pdf(pos=[cq, eta])
        """
        delta_z = (pos[0][1] - pos[0][0]) / 2
        delta_e = (pos[1][1] - pos[1][0]) / 2
        x = [pos[0][0] - delta_z, pos[0][-1] + delta_z]
        y = [pos[1][0] - delta_e, pos[1][-1] + delta_e]

        x_size = pos[0].size
        y_size = pos[1].size
        zeta, eta = self.rvs(size)
        hist, _, _ = np.histogram2d(zeta, eta, bins=[x_size, y_size], range=[x, y])

        hist /= hist.sum()

        x_, y_ = np.meshgrid(pos[0], pos[1])
        return x_, y_, hist.T


class CzjzekDistribution(AbstractDistribution):
    r"""A Czjzek distribution model class.

    The Czjzek distribution model is a random sampling of second-rank traceless
    symmetric tensors whose explicit matrix form follows

    .. math::
        {\bf S} = \left[
        \begin{array}{l l l}
        \sqrt{3} U_5 - U_1   & \sqrt{3} U_4          & \sqrt{3} U_2 \\
        \sqrt{3} U_4         & -\sqrt{3} U_5 - U_1   & \sqrt{3} U_3 \\
        \sqrt{3} U_2         & \sqrt{3} U_3          & 2 U_1
        \end{array}
        \right],

    where the components, :math:`U_i`, are randomly drawn from a five-dimensional
    multivariate normal distribution. Each component, :math:`U_i`, is a dimension of
    the five-dimensional uncorrelated multivariate normal distribution with the mean
    of :math:`<U_i>=0` and the variance :math:`<U_iU_i>=\sigma^2`.

    .. math::
        S_T = S_C(\sigma),

    Args:
        float sigma: The Gaussian standard deviation.

    .. note:: In the original Czjzek paper, the parameter :math:`\sigma` is given as
        two times the standard deviation of the multi-variate normal distribution used
        here.

    Example:
        >>> from mrsimulator.models import CzjzekDistribution
        >>> cz_model = CzjzekDistribution(0.5)
    """

    def __init__(self, sigma: float, polar=False):
        self.sigma = sigma
        self.polar = polar

    def rvs(self, size: int):
        """Draw random variates of length `size` from the distribution.

        Args:
            size: The number of random points to draw.

        Returns:
            A list of two NumPy array, where the first and the second array are the
            anisotropic/quadrupolar coupling constant and asymmetry parameter,
            respectively.

        Example:
            >>> Cq_dist, eta_dist = cz_model.rvs(size=1000000)
        """
        tensors = _czjzek_random_distribution_tensors(self.sigma, size)
        if not self.polar:
            return get_Haeberlen_components(tensors)
        return x_y_from_zeta_eta(*get_Haeberlen_components(tensors))


class ExtCzjzekDistribution(AbstractDistribution):
    r"""An extended Czjzek distribution distribution model.

    The extended Czjzek random distribution [#f1]_ model is an extension of the Czjzek
    model, given as

    .. math::
        S_T = S(0) + \rho S_C(\sigma=1),

    where :math:`S_T` is the total tensor, :math:`S(0)` is the dominant tensor,
    :math:`S_C(\sigma=1)` is the Czjzek random model attributing to the random
    perturbation of the tensor about the dominant tensor, :math:`S(0)`, and
    :math:`\rho` is the size of the perturbation. Note, in the above equation, the
    :math:`\sigma` parameter from the Czjzek random model, :math:`S_C`, has no meaning
    and is set to one. The factor, :math:`\rho`, is defined as

    .. math::
        \rho = \frac{||S(0)|| \epsilon}{\sqrt{30}},

    where :math:`\|S(0)\|` is the 2-norm of the dominant tensor, and :math:`\epsilon`
    is a fraction.

    .. [#f1] Gérard Le Caër, Bruno Bureau, and Dominique Massiot,
        An extension of the Czjzek model for the distributions of electric field
        gradients in disordered solids and an application to NMR spectra of 71Ga in
        chalcogenide glasses. Journal of Physics: Condensed Matter, 2010, 22, 065402.
        DOI: 10.1088/0953-8984/22/6/065402

    Args:
        SymmetricTensor symmetric_tensor: A shielding or quadrupolar symmetric tensor
            or equivalent dict object.
        float eps: A fraction determining the extent of perturbation.

    Example
    -------

    >>> from mrsimulator.models import ExtCzjzekDistribution
    >>> S0 = {"Cq": 1e6, "eta": 0.3}
    >>> ext_cz_model = ExtCzjzekDistribution(S0, eps=0.35)
    """

    def __init__(self, symmetric_tensor: SymmetricTensor, eps: float, polar=False):
        self.symmetric_tensor = symmetric_tensor
        self.eps = eps
        self.polar = polar

    def rvs(self, size: int):
        """Draw random variates of length `size` from the distribution.

        Args:
            size: The number of random points to draw.

        Returns:
            A list of two NumPy array, where the first and the second array are the
            anisotropic/quadrupolar coupling constant and asymmetry parameter,
            respectively.

        Example:
            >>> Cq_dist, eta_dist = ext_cz_model.rvs(size=1000000)
        """

        # czjzek_random_distribution model
        tensors = _czjzek_random_distribution_tensors(1, size)

        symmetric_tensor = self.symmetric_tensor

        if isinstance(symmetric_tensor, dict):
            symmetric_tensor = SymmetricTensor(**symmetric_tensor)

        zeta = symmetric_tensor.zeta or symmetric_tensor.Cq
        eta = symmetric_tensor.eta

        # the traceless second-rank symmetric Cartesian tensor in PAS
        T0 = [0.0, 0.0, 0.0]
        if zeta != 0 and eta != 0:
            T0 = get_principal_components(zeta, eta)

        # 2-norm of the tensor
        norm_T0 = np.linalg.norm(T0)

        # the perturbation factor
        rho = self.eps * norm_T0 / np.sqrt(30)

        # total tensor
        total_tensors = np.diag(T0) + rho * tensors

        if not self.polar:
            return get_Haeberlen_components(total_tensors)
        return x_y_from_zeta_eta(*get_Haeberlen_components(total_tensors))



# -*- coding: utf-8 -*-
import os
import timeit

import numpy as np
from mrsimulator import __version__
from mrsimulator import Simulator
from mrsimulator import Site
from mrsimulator import SpinSystem
from mrsimulator.methods import BlochDecayCentralTransitionSpectrum
from mrsimulator.methods import BlochDecaySpectrum

# import platform
# os_system = platform.system()


def generate_spin_half_spin_system(n=1000):
    iso = np.random.normal(loc=0.0, scale=10.0, size=n)
    zeta = np.random.normal(loc=50.0, scale=15.12, size=n)
    eta = np.random.normal(loc=0.25, scale=0.012, size=n)
    spin_systems = []
    for i, z, e in zip(iso, zeta, eta):
        site = Site(
            isotope="29Si",
            isotropic_chemical_shift=i,
            shielding_symmetric={"zeta": z, "eta": e},
        )
        spin_systems.append(SpinSystem(sites=[site]))
    return spin_systems


def generate_spin_half_int_quad_spin_system(n=1000):
    iso = np.random.normal(loc=0.0, scale=10.0, size=n)
    Cq = np.random.normal(loc=5.0e6, scale=1e5, size=n)
    eta = np.random.normal(loc=0.1, scale=0.012, size=n)
    spin_systems = []
    for i, c, e in zip(iso, Cq, eta):
        site = Site(
            isotope="17O", isotropic_chemical_shift=i, quadrupolar={"Cq": c, "eta": e}
        )
        spin_systems.append(SpinSystem(sites=[site]))
    return spin_systems


def generate_simulator(spin_systems, method):
    sim = Simulator()
    sim.spin_systems = spin_systems
    sim.methods = [method]
    return sim


def execute(sim, n_jobs=1):
    sim.run(n_jobs=n_jobs)


def CSA_static_method():
    return BlochDecaySpectrum(
        channels=["29Si"], spectral_dimensions=[dict(spectral_width=25000)]
    )


def CSA_MAS_method():
    return BlochDecaySpectrum(
        channels=["29Si"],
        rotor_frequency=5000,
        spectral_dimensions=[dict(spectral_width=25000)],
    )


def CSA_VAS_method():
    return BlochDecaySpectrum(
        channels=["29Si"],
        rotor_frequency=5000,
        rotor_angle=1.57079,
        spectral_dimensions=[dict(spectral_width=25000)],
    )


def quad_static_method():
    return BlochDecayCentralTransitionSpectrum(
        channels=["17O"], spectral_dimensions=[{"spectral_width": 50000}]
    )


def quad_MAS_method():
    return BlochDecayCentralTransitionSpectrum(
        channels=["17O"],
        rotor_frequency=14000,
        spectral_dimensions=[{"spectral_width": 50000}],
    )


def time_string(time):
    count = 0
    if time < 0.003:  # i.e 4 ms
        color = "\033[92m"
    if time >= 0.003 and time < 0.005:
        color = "\033[93m"
    if time >= 0.005:
        color = "\033[91m"

    # color = '' if os_system == 'Windows' else color

    while time < 1:
        time *= 1000  # ms
        count += 1
    if count == 0:
        return f"{time:.3f}  s", color
    if count == 1:
        return f"{time:.3f} ms", color
    if count == 2:
        return f"{time:.3f} µs", color


def blocks(n_blocks, fn_string, n, level, n_jobs=1):
    description = [
        "Static CSA only spectrum",
        "Static quadrupolar (1st + 2nd order) only spectrum",
        "MAS CSA only sidebands spectrum",
        "VAS CSA only spectrum",
        "MAS quadrupolar (1st + 2nd order) only spectrum",
    ]
    print(f"\nLevel {level} results.")
    print("Average computation time for simulation one single-site spin system.")
    print(
        f"Reported value is the simulation time per spectra averaged over {n} spectra "
        "generated for random tensor parameters."
    )
    size = os.get_terminal_size().columns
    delmit = "-"
    print(f"{delmit:-<{size}}")
    left_align = "Computation method"
    right_align = "Average time"
    print(f"{left_align:<{size-15}}{right_align:>15}")
    print(f"{delmit:-<{size}}")
    for i in range(n_blocks):
        t = timeit.timeit(
            f"execute(sim[{i}], {n_jobs})",
            setup=f"sim={fn_string}({n})",
            globals=globals(),
            number=1,
        )
        t /= n  # s
        t, color = time_string(t)
        end = "\033[0m"
        # end = '' if os_system == 'Windows' else "\033[0m"
        size = os.get_terminal_size().columns - 15
        print(f"{end}{description[i]:.<{size}}{color}{t:>15}{end}")


def level_n_CSA_static(n):
    # spin 1/2
    spin_sys = generate_spin_half_spin_system(n)

    method = CSA_static_method()  # static
    sim_csa_static = generate_simulator(spin_sys, method)

    method = CSA_MAS_method()  # mas
    sim_csa_mas = generate_simulator(spin_sys, method)

    method = CSA_VAS_method()  # vas
    sim_csa_vas = generate_simulator(spin_sys, method)

    # quad
    spin_sys = generate_spin_half_int_quad_spin_system(n)

    method = quad_static_method()  # static
    sim_quad_static = generate_simulator(spin_sys, method)

    method = quad_MAS_method()  # static
    sim_quad_mas = generate_simulator(spin_sys, method)

    return sim_csa_static, sim_quad_static, sim_csa_mas, sim_csa_vas, sim_quad_mas


class Benchmark:
    @staticmethod
    def prep():
        print(f"Benchmarking using mrsimulator version {__version__}")

    @staticmethod
    def l0(n_jobs):
        blocks(5, "level_n_CSA_static", 10, 0, n_jobs)

    @staticmethod
    def l1(n_jobs):
        blocks(5, "level_n_CSA_static", 2000, 1, n_jobs)

    @staticmethod
    def l2(n_jobs):
        blocks(5, "level_n_CSA_static", 10000, 2, n_jobs)



# -*- coding: utf-8 -*-
"""The Event class."""
from sys import modules
from typing import List
from typing import Union

import csdmpy as cp
from pydantic import BaseModel

from . import affine as af  # noqa:F401
from . import apodization as ap  # noqa:F401
from . import baseline as bl  # noqa:F401
from ._base import Operation


__author__ = "Maxwell C. Venetos"
__email__ = "maxvenetos@gmail.com"


class SignalProcessor(BaseModel):
    """Signal processing class to apply a series of operations to the dependent
    variables of the simulation dataset.

    Attributes
    ----------

    operations: List
        A list of operations.

    Examples
    --------

    >>> post_sim = SignalProcessor(operations=[o1, o2]) # doctest: +SKIP
    """

    processed_data: cp.CSDM = None
    operations: List[Operation] = []

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    @classmethod
    def parse_dict_with_units(cls, py_dict: dict):
        """Parse a list of operations dictionary to a SignalProcessor class object.

        Args:
            pt_dict: A python dict object.
        """
        mod = modules[__name__]
        lst = [
            getattr(getattr(mod, op["function"]), op["type"]).parse_dict_with_units(op)
            if "type" in op
            else getattr(mod, op["function"]).parse_dict_with_units(op)
            for op in py_dict["operations"]
        ]
        return SignalProcessor(operations=lst)

    def json(self) -> dict:
        """Parse the class object to a JSON compliant python dictionary object, where
        the attribute value with physical quantity is expressed as a string with a
        value and a unit.

        Returns:
            A Dict object.
        """
        op = {}
        op["operations"] = [item.json() for item in self.operations]
        return op

    def apply_operations(self, data, **kwargs):
        """Function to apply all the operation functions in the operations member of a
        SignalProcessor object. Operations applied sequentially over the data member.

        Returns:
            CSDM object: A copy of the data member with the operations applied to it.
        """
        if not isinstance(data, cp.CSDM):
            raise ValueError("The data must be a CSDM object.")
        for filters in self.operations:
            data = filters.operate(data)
        self.processed_data = data

        return data


class Scale(Operation):
    r"""Scale the amplitudes of all dependent variables (y) from a CSDM object.

    .. math::
        f(y) = \text{factor} \times y

    Args:
        float factor: The scaling factor. The default value is 1.

    Example
    -------

    >>> from mrsimulator import signal_processing as sp
    >>> operation1 = sp.Scale(factor=20)
    """

    factor: float = 1

    def operate(self, data):
        """Applies the operation.

        Args:
            data: CSDM object
        """
        data *= self.factor
        return data


class Linear(Operation):
    r"""Apply linear operation across all dependent variables (y) from a CSDM object.

    .. math::
            f(y) = \text{amplitude} \times y + \text{offset}

    Args:
        float amplitude: The scaling factor. The default value is 1.
        float offsett: The offset factor. The default value is 0.

    Example
    -------

    >>> from mrsimulator import signal_processing as sp
    >>> operation1 = sp.Linear(amplitude=20, offset=-10)
    """

    amplitude: float = 1
    offset: float = 0

    def operate(self, data):
        """Applies the operation.

        Args:
            data: CSDM object
        """
        data *= self.amplitude
        data += self.offset
        return data


class IFFT(Operation):
    """Apply an inverse Fourier transform on all dependent variables of the CSDM object.

    Args:
        int dim_index: Dimension index along which the function is applied.

    Example
    -------

    >>> operation2 = sp.IFFT(dim_index=0)
    """

    dim_index: Union[int, list, tuple] = 0

    def operate(self, data):
        """Applies the operation.

        Args:
            data: CSDM object
        """
        d_i = [self.dim_index] if isinstance(self.dim_index, int) else self.dim_index
        for i in d_i:
            data = data.fft(axis=i)
        return data


class FFT(IFFT):
    """Apply a forward Fourier transform on all dependent variables of the CSDM object.

    Args:
        int dim_index: Dimension index along which the function is applied.

    Example
    -------

    >>> operation3 = sp.FFT(dim_index=0)
    """


class complex_conjugate(Operation):
    pass



# -*- coding: utf-8 -*-
"""
Function list:
    - ThreeQ_VAS
    - FiveQ_VAS
    - SevenQ_VAS
"""
from .base import BaseNamedMethod2D

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


shear_factor_MQ_MAS = {
    3: {1.5: 21 / 27, 2.5: 114 / 72, 3.5: 303 / 135, 4.5: 546 / 216},
    5: {2.5: 150 / 72, 3.5: 165 / 135, 4.5: 570 / 216},
    7: {3.5: 483 / 135, 4.5: 84 / 216},
    9: {4.5: 1116 / 216},
}

MQ_p_symmetry = {
    "ThreeQ_VAS": {"mq": 1.5},
    "FiveQ_VAS": {"mq": 2.5},
    "SevenQ_VAS": {"mq": 3.5},
}


class MQ_VAS(BaseNamedMethod2D):
    r"""A generic multiple-quantum variable-angle spinning method for simulating average
    frequencies. The resulting spectrum is sheared such that the correlating dimensions
    are the isotropic dimension and the VAS dimension, respectively, where the isotropic
    dimension is given as

    .. math::
        \nu_\text{MQ-iso} = 1/(1+k) \nu_\text{MQ}(-m, m) + k/(1+k) \nu_\text{MAS}

    Args:
        channels: A list of isotope symbols over which the method will be applied.
        spectral_dimensions: A list of python dict. Each dict is contains keywords that
            describe the coordinates along a spectral dimension. The keywords along with
            its definition are:

            - count:
                An optional integer with the number of points, :math:`N`, along the
                dimension. The default value is 1024.
            - spectral_width:
                An `optional` float with the spectral width, :math:`\Delta x`, along the
                dimension in units of Hz. The default is 25 kHz.
            - reference_offset:
                An `optional` float with the reference offset, :math:`x_0` along the
                dimension in units of Hz. The default value is 0 Hz.
            - origin_offset:
                An `optional` float with the origin offset (Larmor frequency) along the
                dimension in units of Hz. The default value is None.

        magetic_flux_density: An `optional` float containing the macroscopic magnetic
            flux density, :math:`H_0`, of the applied external magnetic field in units
            of T. The default value is 9.4.
        rotor_angle: An `optional` float containing the angle between the sample
            rotation axis and the applied external magnetic field, :math:`\theta`, in
            units of rad. The default value is 0.9553166, i.e. the magic angle.

    note:
        The `rotor_frequency` parameter is fixed for this method. The method produces
        an infinite spinning speed spectrum.

    Return:
        A :class:`~mrsimulator.Method` instance.
    """

    def __new__(cls, **kwargs):
        return cls.update(super().__new__(cls, **kwargs))

    @staticmethod
    def update(method):
        mq = MQ_p_symmetry[method.name]["mq"]
        spin = method.channels[0].spin

        # select the coherence for the first event
        P = int(2 * mq)
        nQ = P
        P = -P if mq == spin else P

        # setting transition symmetry elements
        sd = method.spectral_dimensions
        sd[0].events[0].transition_query.P = {"channel-1": [[P]]}
        sd[0].events[0].transition_query.D = {"channel-1": [[0]]}

        sd[1].events[0].transition_query.P = {"channel-1": [[-1]]}
        sd[1].events[0].transition_query.D = {"channel-1": [[0]]}

        # method affine matrix
        if method.affine_matrix is None:
            k = shear_factor_MQ_MAS[nQ][spin]
            method.affine_matrix = [1 / (1 + k), k / (1 + k), 0, 1]

        # method description
        method.description = f"Simulate a {nQ}Q variable-angle spinning spectrum."
        return method


class ThreeQ_VAS(MQ_VAS):
    r"""Simulate a sheared and scaled 3Q 2D variable-angle spinning spectrum.

    Args:
        channels: A list of isotope symbols over which the method will be applied.
        spectral_dimensions: A list of python dict. Each dict is contains keywords that
            describe the coordinates along a spectral dimension. The keywords along with
            its definition are:

            - count:
                An optional integer with the number of points, :math:`N`, along the
                dimension. The default value is 1024.
            - spectral_width:
                An `optional` float with the spectral width, :math:`\Delta x`, along the
                dimension in units of Hz. The default is 25 kHz.
            - reference_offset:
                An `optional` float with the reference offset, :math:`x_0` along the
                dimension in units of Hz. The default value is 0 Hz.
            - origin_offset:
                An `optional` float with the origin offset (Larmor frequency) along the
                dimension in units of Hz. The default value is None.

        magetic_flux_density: An `optional` float containing the macroscopic magnetic
            flux density, :math:`H_0`, of the applied external magnetic field in units
            of T. The default value is 9.4.
        rotor_angle: An `optional` float containing the angle between the sample
            rotation axis and the applied external magnetic field, :math:`\theta`, in
            units of rad. The default value is 0.9553166, i.e. the magic angle.

    Note:
        The attribute `rotor_frequency` cannot be modified for this method and is set to
        simulate an infinite speed spectrum.

    Return:
        A :class:`~mrsimulator.Method` instance.

    Example:
        >>> method = ThreeQ_VAS(
        ...     channels=["87Rb"],
        ...     magnetic_flux_density=7,  # in T
        ...     spectral_dimensions=[
        ...         {
        ...             "count": 128,
        ...             "spectral_width": 3e3,  # in Hz
        ...             "reference_offset": -2e3,  # in Hz
        ...             "label": "Isotropic dimension",
        ...         },
        ...         {
        ...             "count": 512,
        ...             "spectral_width": 1e4,  # in Hz
        ...             "reference_offset": -5e3,  # in Hz
        ...             "label": "MAS dimension",
        ...         },
        ...     ],
        ... )
        >>> sys = SpinSystem(sites=[Site(isotope='87Rb')])
        >>> method.get_transition_pathways(sys)
        [|-1.5⟩⟨1.5| ⟶ |-0.5⟩⟨0.5|]
    """


class FiveQ_VAS(MQ_VAS):
    r"""Simulate a sheared and scaled 5Q variable-angle spinning spectrum.

    Args:
        channels: A list of isotope symbols over which the method will be applied.
        spectral_dimensions: A list of python dict. Each dict is contains keywords that
            describe the coordinates along a spectral dimension. The keywords along with
            its definition are:

            - count:
                An optional integer with the number of points, :math:`N`, along the
                dimension. The default value is 1024.
            - spectral_width:
                An `optional` float with the spectral width, :math:`\Delta x`, along the
                dimension in units of Hz. The default is 25 kHz.
            - reference_offset:
                An `optional` float with the reference offset, :math:`x_0` along the
                dimension in units of Hz. The default value is 0 Hz.
            - origin_offset:
                An `optional` float with the origin offset (Larmor frequency) along the
                dimension in units of Hz. The default value is None.

        magetic_flux_density: An `optional` float containing the macroscopic magnetic
            flux density, :math:`H_0`, of the applied external magnetic field in units
            of T. The default value is 9.4.
        rotor_angle: An `optional` float containing the angle between the sample
            rotation axis and the applied external magnetic field, :math:`\theta`, in
            units of rad. The default value is 0.9553166, i.e. the magic angle.

    Note:
        The attribute `rotor_frequency` cannot be modified for this method and is set to
        simulate an infinite speed spectrum.

    Return:
        A :class:`~mrsimulator.Method` instance.

    Example:
        >>> method = FiveQ_VAS(
        ...     channels=["17O"],
        ...     magnetic_flux_density=11.7,  # in T
        ...     spectral_dimensions=[
        ...         {
        ...             "count": 512,
        ...             "spectral_width": 5e3,  # in Hz
        ...             "reference_offset": -3e3,  # in Hz
        ...             "label": "Isotropic dimension",
        ...         },
        ...         {
        ...             "count": 512,
        ...             "spectral_width": 2e4,  # in Hz
        ...             "reference_offset": -2e3,  # in Hz
        ...             "label": "MAS dimension",
        ...         },
        ...     ],
        ... )
        >>> sys = SpinSystem(sites=[Site(isotope='17O')])
        >>> method.get_transition_pathways(sys)
        [|-2.5⟩⟨2.5| ⟶ |-0.5⟩⟨0.5|]
    """


class SevenQ_VAS(MQ_VAS):
    r"""Simulate a sheared and scaled 7Q variable-angle spinning spectrum.

    Args:
        channels: A list of isotope symbols over which the method will be applied.
        spectral_dimensions: A list of python dict. Each dict is contains keywords that
            describe the coordinates along a spectral dimension. The keywords along with
            its definition are:

            - count:
                An optional integer with the number of points, :math:`N`, along the
                dimension. The default value is 1024.
            - spectral_width:
                An `optional` float with the spectral width, :math:`\Delta x`, along the
                dimension in units of Hz. The default is 25 kHz.
            - reference_offset:
                An `optional` float with the reference offset, :math:`x_0` along the
                dimension in units of Hz. The default value is 0 Hz.
            - origin_offset:
                An `optional` float with the origin offset (Larmor frequency) along the
                dimension in units of Hz. The default value is None.

        magetic_flux_density: An `optional` float containing the macroscopic magnetic
            flux density, :math:`H_0`, of the applied external magnetic field in units
            of T. The default value is 9.4.
        rotor_angle: An `optional` float containing the angle between the sample
            rotation axis and the applied external magnetic field, :math:`\theta`, in
            units of rad. The default value is 0.9553166, i.e. the magic angle.

    Note:
        The attribute `rotor_frequency` cannot be modified for this method and is set to
        simulate an infinite speed spectrum.

    Return:
        A :class:`~mrsimulator.Method` instance.

    Example:
        >>> method = SevenQ_VAS(
        ...     channels=["51V"],
        ...     magnetic_flux_density=4.7,  # in T
        ...     spectral_dimensions=[
        ...         {
        ...             "count": 256,
        ...             "spectral_width": 1e3,  # in Hz
        ...             "reference_offset": 1e3,  # in Hz
        ...             "label": "Isotropic dimension",
        ...         },
        ...         {
        ...             "count": 1024,
        ...             "spectral_width": 1e4,  # in Hz
        ...             "reference_offset": -2e3,  # in Hz
        ...             "label": "MAS dimension",
        ...         },
        ...     ],
        ... )
        >>> sys = SpinSystem(sites=[Site(isotope='51V')])
        >>> method.get_transition_pathways(sys)
        [|-3.5⟩⟨3.5| ⟶ |-0.5⟩⟨0.5|]
    """



# -*- coding: utf-8 -*-
import numpy as np
import pytest
from mrsimulator.method.transition_query import TransitionQuery
from mrsimulator.methods import SSB2D

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


def test_SSB_rotor_freq():
    e = "`rotor_frequency` cannot be zero for SSB2D method."
    with pytest.raises(ValueError, match=f".*{e}.*"):
        SSB2D(spectral_dimensions=[{}, {}])

    with pytest.raises(ValueError, match=f".*{e}.*"):
        SSB2D(rotor_frequency=0, spectral_dimensions=[{}, {}])


def test_spectral_dimension_count():
    e = "Method requires exactly 2 spectral dimensions, given 1."
    with pytest.raises(ValueError, match=f".*{e}.*"):
        SSB2D(rotor_frequency=1e4, spectral_dimensions=[{}])


def test_SSB_setting_events():
    e = "`events` value cannot be modified for SSB2D method."
    with pytest.raises(ValueError, match=f".*{e}.*"):
        SSB2D(rotor_frequency=1e4, spectral_dimensions=[{"events": [{}]}, {}])


def test_SSB_affine():
    mth = SSB2D(channels=["13C"], rotor_frequency=1200)
    np.allclose(mth.affine_matrix, [1, -1, 0, 0])
    assert SSB2D.parse_dict_with_units(mth.json()) == mth


def test_SSB_general():
    """Inner satellite-transition variable-angle spinning method"""
    mth = SSB2D(
        channels=["87Rb"],
        magnetic_flux_density=9.4,  # in T
        rotor_frequency=1200,
        spectral_dimensions=[
            {"count": 1024, "spectral_width": 5e4},
            {"count": 1024, "spectral_width": 5e4},
        ],
    )
    assert mth.name == "SSB2D"

    assert mth.description == "Simulate a 2D sideband separation method."

    # test transition query
    tq = TransitionQuery(P={"channel-1": [[-1]]}, D={"channel-1": [[0]]})
    assert mth.spectral_dimensions[0].events[0].transition_query == tq
    assert mth.spectral_dimensions[1].events[0].transition_query == tq

    # test rotor_frequency
    assert mth.spectral_dimensions[0].events[0].rotor_frequency == 1200
    assert mth.spectral_dimensions[1].events[0].rotor_frequency == 1e12

    # check serialization
    assert SSB2D.parse_dict_with_units(mth.json()) == mth

    assert np.allclose(mth.affine_matrix, [1, -1, 0.0, 1.0])

    serialize = mth.json()
    _ = serialize.pop("affine_matrix")

    assert serialize == {
        "channels": ["87Rb"],
        "description": mth.description,
        "magnetic_flux_density": "9.4 T",
        "name": "SSB2D",
        "rotor_angle": "0.955316618 rad",
        "rotor_frequency": "1200.0 Hz",
        "spectral_dimensions": [
            {
                "count": 1024,
                "reference_offset": "0.0 Hz",
                "spectral_width": "50000.0 Hz",
            },
            {
                "count": 1024,
                "events": [{"rotor_frequency": "1000000000000.0 Hz"}],
                "reference_offset": "0.0 Hz",
                "spectral_width": "50000.0 Hz",
            },
        ],
    }



# -*- coding: utf-8 -*-
"""
Function list:
    - ST1_VAS
    - ST2_VAS
"""
from .base import BaseNamedMethod2D

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


shear_factor_ST_MAS = {
    3: {1.5: 24 / 27, 2.5: 21 / 72, 3.5: 84 / 135, 4.5: 165 / 216},
    5: {2.5: 132 / 72, 3.5: 69 / 135, 4.5: 12 / 216},
    7: {3.5: 324 / 135, 4.5: 243 / 216},
    9: {4.5: 600 / 216},
}

ST_p_symmetry = {"ST1_VAS": {"st": 1.5}, "ST2_VAS": {"st": 2.5}}


class ST_VAS(BaseNamedMethod2D):
    r"""Simulate a satellite-transition magic-angle spinning spectrum.
    Args:
        channels: A list of isotope symbols over which the method will be applied.
        spectral_dimensions: A list of python dict. Each dict is contains keywords that
            describe the coordinates along a spectral dimension. The keywords along with
            its definition are:

            - count:
                An optional integer with the number of points, :math:`N`, along the
                dimension. The default value is 1024.
            - spectral_width:
                An `optional` float with the spectral width, :math:`\Delta x`, along the
                dimension in units of Hz. The default is 25 kHz.
            - reference_offset:
                An `optional` float with the reference offset, :math:`x_0` along the
                dimension in units of Hz. The default value is 0 Hz.
            - origin_offset:
                An `optional` float with the origin offset (Larmor frequency) along the
                dimension in units of Hz. The default value is None.
        magetic_flux_density: An `optional` float containing the macroscopic magnetic
            flux density, :math:`H_0`, of the applied external magnetic field in units
            of T. The default value is ``9.4``.

    note:
        The `rotor_frequency` and `rotor_angle` parameters are fixed for this method.
        The method produces an infinite spinning speed spectrum.

    Return:
        A :class:`~mrsimulator.Method` instance.
    """

    def __new__(cls, **kwargs):
        return cls.update(super().__new__(cls, **kwargs))

    @staticmethod
    def update(method):
        st = ST_p_symmetry[method.name]["st"]
        spin = method.channels[0].spin

        # select the coherence for the first event
        d = st ** 2 - (st - 1) ** 2
        D = [[d], [-d]]

        # setting transition symmetry elements
        sd = method.spectral_dimensions
        sd[0].events[0].transition_query.P = {"channel-1": [[-1]]}
        sd[0].events[0].transition_query.D = {"channel-1": D}

        sd[1].events[0].transition_query.P = {"channel-1": [[-1]]}
        sd[1].events[0].transition_query.D = {"channel-1": [[0]]}

        # method affine matrix
        if method.affine_matrix is None:
            k = shear_factor_ST_MAS[int(2 * st)][spin]
            sign = 1  # if st == spin else -1
            method.affine_matrix = [1 / (1 + k), sign * k / (1 + k), 0, 1]

        # method description
        method.description = (
            f"Simulate a {st} -> {st-1} and {-st+1} -> {-st} satellite-transition "
            "variable-angle spinning spectrum."
        )
        return method


class ST1_VAS(ST_VAS):
    r"""Simulate a sheared and scaled inner satellite and central transition correlation
    spectrum.

    Args:
        channels: A list of isotope symbols over which the method will be applied.
        spectral_dimensions: A list of python dict. Each dict is contains keywords that
            describe the coordinates along a spectral dimension. The keywords along with
            its definition are:

            - count:
                An optional integer with the number of points, :math:`N`, along the
                dimension. The default value is 1024.
            - spectral_width:
                An `optional` float with the spectral width, :math:`\Delta x`, along the
                dimension in units of Hz. The default is 25 kHz.
            - reference_offset:
                An `optional` float with the reference offset, :math:`x_0` along the
                dimension in units of Hz. The default value is 0 Hz.
            - origin_offset:
                An `optional` float with the origin offset (Larmor frequency) along the
                dimension in units of Hz. The default value is None.

        magetic_flux_density: An `optional` float containing the macroscopic magnetic
            flux density, :math:`H_0`, of the applied external magnetic field in units
            of T. The default value is 9.4.
        rotor_angle: An `optional` float containing the angle between the sample
            rotation axis and the applied external magnetic field, :math:`\theta`, in
            units of rad. The default value is 0.9553166, i.e. the magic angle.

    Note:
        The attribute `rotor_frequency` cannot be modified for this method and is set to
        simulate an infinite speed spectrum.

    Return:
        A :class:`~mrsimulator.Method` instance.

    Example:
        >>> method = ST1_VAS(
        ...     channels=["87Rb"],
        ...     magnetic_flux_density=9.4,  # in T
        ...     spectral_dimensions=[
        ...         {
        ...             "count": 128,
        ...             "spectral_width": 1e3,  # in Hz
        ...             "reference_offset": -5e3,  # in Hz
        ...             "label": "Isotropic dimension",
        ...         },
        ...         {
        ...             "count": 256,
        ...             "spectral_width": 1e4,  # in Hz
        ...             "reference_offset": -3e3,  # in Hz
        ...             "label": "MAS dimension",
        ...         },
        ...     ],
        ... )
        >>> sys = SpinSystem(sites=[Site(isotope='87Rb')])
        >>> pprint(method.get_transition_pathways(sys))
        [|-1.5⟩⟨-0.5| ⟶ |-0.5⟩⟨0.5|, |0.5⟩⟨1.5| ⟶ |-0.5⟩⟨0.5|]
    """


class ST2_VAS(ST_VAS):
    r"""Simulate a sheared and scaled second to inner satellite and central transition
    correlation spectrum.

    Args:
        channels: A list of isotope symbols over which the method will be applied.
        spectral_dimensions: A list of python dict. Each dict is contains keywords that
            describe the coordinates along a spectral dimension. The keywords along with
            its definition are:

            - count:
                An optional integer with the number of points, :math:`N`, along the
                dimension. The default value is 1024.
            - spectral_width:
                An `optional` float with the spectral width, :math:`\Delta x`, along the
                dimension in units of Hz. The default is 25 kHz.
            - reference_offset:
                An `optional` float with the reference offset, :math:`x_0` along the
                dimension in units of Hz. The default value is 0 Hz.
            - origin_offset:
                An `optional` float with the origin offset (Larmor frequency) along the
                dimension in units of Hz. The default value is None.

        magetic_flux_density: An `optional` float containing the macroscopic magnetic
            flux density, :math:`H_0`, of the applied external magnetic field in units
            of T. The default value is 9.4.
        rotor_angle: An `optional` float containing the angle between the sample
            rotation axis and the applied external magnetic field, :math:`\theta`, in
            units of rad. The default value is 0.9553166, i.e. the magic angle.

    Note:
        The attribute `rotor_frequency` cannot be modified for this method and is set to
        simulate an infinite speed spectrum.

    Return:
        A :class:`~mrsimulator.Method` instance.

    Example:
        >>> method = ST2_VAS(
        ...     channels=["17O"],
        ...     magnetic_flux_density=9.4,  # in T
        ...     spectral_dimensions=[
        ...         {
        ...             "count": 256,
        ...             "spectral_width": 4e3,  # in Hz
        ...             "reference_offset": -5e3,  # in Hz
        ...             "label": "Isotropic dimension",
        ...         },
        ...         {
        ...             "count": 512,
        ...             "spectral_width": 1e4,  # in Hz
        ...             "reference_offset": -4e3,  # in Hz
        ...             "label": "MAS dimension",
        ...         },
        ...     ],
        ... )
        >>> sys = SpinSystem(sites=[Site(isotope='17O')])
        >>> pprint(method.get_transition_pathways(sys))
        [|-2.5⟩⟨-1.5| ⟶ |-0.5⟩⟨0.5|, |1.5⟩⟨2.5| ⟶ |-0.5⟩⟨0.5|]
    """
